# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module tree_based_tensor_learning.

'''

from copy import deepcopy
from itertools import combinations
from random import shuffle
import numpy as np
import tensap


class TreeBasedTensorLearning(tensap.TensorLearning):
    '''
    Class TreeBasedTensorLearning.

    See also tensap.TensorLearning.

    Attributes
    ----------
    tree : tensap.DimensionTree
        The dimension tree of the tree-based tensor.
    is_active_node : list or numpy.ndarray
        Booleans indicating if each node is active.

    '''

    def __init__(self, tree, is_active_node, *args):
        '''
        Constructor for the class TreeBasedTensorLearning.

        See also tensap.TensorLearning.

        Parameters
        ----------
        tree : tensap.DimensionTree
            The dimension tree of the tree-based tensor.
        is_active_node : list or numpy.ndarray
            Booleans indicating if each node is active.
        *args : tuple
            Additional parameters.

        Returns
        -------
        None.

        '''
        super().__init__(*args)

        self.tree = tree
        self.is_active_node = list(map(bool, is_active_node))
        self.order = tree.dim2ind.size
        self._number_of_parameters = np.count_nonzero(is_active_node)

        self.initialization_type = 'canonical'
        self.rank_adaptation_options['rank_one_correction'] = True
        self.rank_adaptation_options['theta'] = 0.8
        self.linear_model_learning_parameters[
            'basis_adaptation_internal_nodes'] = False

# %% Standard solver methods
    def initialize(self):
        assert self.tree is not None, \
            'Must provide a DimensionTree object in property tree.'

        if np.isscalar(self.rank) or len(self.rank) == self.order:
            rank = np.zeros(self.tree.nb_nodes, dtype=int)
            rank[self.is_active_node] = self.rank
            self.rank = rank
        self.rank[self.tree.root-1] = self.output_dimension

        shape = [x.shape[1] for x in self.bases_eval]
        if self.initialization_type == 'random':
            f = tensap.TreeBasedTensor.randn(self.tree, self.rank, shape,
                                             self.is_active_node)
        elif self.initialization_type == 'ones':
            f = tensap.TreeBasedTensor.ones(self.tree, self.rank, shape,
                                            self.is_active_node)
        elif self.initialization_type == 'initial_guess':
            f = self.initial_guess
            if not np.all(f.ranks == self.rank):
                tr = tensap.Truncator(tolerance=np.finfo(float).eps,
                                      max_rank=self.rank)
                f = tr.truncate(f)
        elif self.initialization_type == 'mean' or \
                self.initialization_type == 'mean_randomized':
            if not np.all(self.rank == 1):
                raise NotImplementedError('Initialization only implemented ' +
                                          'if np.all(self.rank == 1).')
            if not isinstance(self.training_data, list) or \
                    (isinstance(self.training_data, list) and
                     len(self.training_data) == 1):
                raise NotImplementedError('Initialization type not ' +
                                          'implemented in unsupervised ' +
                                          'learning.')
            if isinstance(self.bases, tensap.FunctionalBases):
                means = self.bases.mean()
            else:
                means = [np.mean(x) for x in self.bases_eval]
            if self.initialization_type == 'mean_randomized':
                means = [x + 0.01*np.random.randn(*x.shape) for x in means]
            means = [tensap.FullTensor(x, 2, [x.shape[0], 1]) for x in means]

            f = tensap.TreeBasedTensor.ones(self.tree, self.rank, shape)
            f.tensors[self.tree.dim2ind-1] = means
            nb_child = len(self.tree.children(self.tree.root))
            if np.ndim(self.training_data[1]) == 2:
                shape = np.concatenate((np.full(nb_child, 1),
                                       [self.training_data[1].shape[1]]))
            else:
                shape = np.full(nb_child, 1)
            f.tensors[self.tree.root-1] = \
                tensap.FullTensor(np.mean(self.training_data[1]), shape=shape)
            f.update_attributes()

            f = f.inactivate_nodes(
                np.nonzero(np.logical_not(self.is_active_node))[0]+1)
        elif self.initialization_type == 'canonical':
            if self.output_dimension != 1:
                print('Canonical initialization not implemented for ' +
                      'outputDimension > 1, performing a random ' +
                      'initialization.')
                f = tensap.TreeBasedTensor.randn(self.tree, self.rank, shape,
                                                 self.is_active_node)
            else:
                f = self.canonical_initialization(np.max(self.rank))
                if not np.all(f.ranks == self.rank):
                    tr = tensap.Truncator(tolerance=np.finfo(float).eps,
                                          max_rank=self.rank)
                    f = tr.truncate(f)
        else:
            raise ValueError('Wrong initialization type.')

        if not np.all(f.ranks == self.rank):
            f = TreeBasedTensorLearning.\
                enriched_edges_to_ranks_random(f, self.rank)

        # Exploration strategy of the tree by increasing level
        tree = f.tree
        exploration_strategy = np.zeros(self._number_of_parameters, dtype=int)
        active_nodes = f.active_nodes
        rep = 0
        for level in range(np.max(tree.level)+1):
            nodes = np.intersect1d(tree.nodes_with_level(level), active_nodes)
            exploration_strategy[rep:rep+len(nodes)] = nodes
            rep += len(nodes)
        self._exploration_strategy = exploration_strategy
        return self, f

    def pre_processing(self, f):
        if isinstance(self.linear_model_learning, (list, np.ndarray)) and \
                len(self.linear_model_learning) != f.tensor.tree.nb_nodes:
            tmp = np.empty(f.tensor.tree.nb_nodes, dtype=object)
            tmp[f.tensor.is_active_node] = self.linear_model_learning
            self.linear_model_learning = tmp
        return self, f

    def randomize_exploration_strategy(self):
        strategy = np.zeros(self._number_of_parameters)
        for level in np.arange(np.max(self.tree.level), -1, -1):
            active_nodes = np.intersect1d(self.tree.nodes_with_level(level),
                                          np.nonzero(self.is_active_node)[0])
            _, ind = np.intersect1d(self._exploration_strategy,
                                    active_nodes,
                                    return_indices=True)
            strategy[ind] = self._exploration_strategy[
                np.random.permutation(ind)]
        return strategy

    def prepare_alternating_minimization_system(self, f, mu):
        tree = f.tensor.tree
        if self.linear_model_learning[mu-1].basis_adaptation:
            if np.isin(mu, tree.internal_nodes):
                if self.linear_model_learning_parameters[
                        'basis_adaptation_internal_nodes']:
                    tr = tensap.Truncator(tolerance=np.finfo(float).eps,
                                          max_rank=np.max(f.tensor.ranks))
                    f.tensor = tr.hsvd(f.tensor)
                elif np.all(f.tensor.is_active_node[tree.children(mu)-1]):
                    self.linear_model_learning[mu-1].basis_adaptation = False
            f.tensor = f.tensor.orth_at_node(mu)
            self.tree = tree
            self.is_active_node = f.tensor.is_active_node
            self.linear_model_learning[mu-1].basis_adaptation_path = \
                self.create_basis_adaptation_path(f.tensor.ranks, mu)
        else:
            f.tensor = f.tensor.orth_at_node(mu)

        grad = f.parameter_gradient_eval(mu)
        if mu == tree.root:
            A = np.reshape(grad.data, [grad.shape[0], -1], order='F')
            self.linear_model_learning[mu-1].initial_guess = np.reshape(
                    f.tensor.tensors[mu-1].data,
                    [-1, f.tensor.ranks[tree.root-1]], order='F')
        else:
            A = np.reshape(grad.data, [grad.shape[0], -1,
                                       f.tensor.ranks[tree.root-1]], order='F')
            self.linear_model_learning[mu-1].initial_guess = np.reshape(
                    f.tensor.tensors[mu-1].data, -1, order='F')

        if isinstance(self.loss_function, tensap.DensityL2LossFunction):
            if isinstance(self.training_data, list) and \
                    len(self.training_data) == 2:
                y = self.training_data[1]
                if isinstance(y, tensap.FunctionalTensor):
                    y = y.tensor
                y = y.orth()
                if tree.is_leaf[mu-1]:
                    a = deepcopy(y)
                    for nod in tree.internal_nodes:
                        a.tensors[nod-1] = tensap.FullTensor(
                            y.tensors[nod-1].data *
                            f.tensor.tensors[nod-1].data,
                            shape=y.tensors[nod-1].shape)
                    ind = np.setdiff1d(range(self.order),
                                       np.nonzero(tree.dim2ind == mu)[0])
                    C = [x.data for x in f.tensor.tensors[tree.dim2ind-1]]
                    b = a.tensor_vector_product([C[x] for x in ind], ind)
                    b = b.tensors[0].data
                else:
                    b = f.tensor.dot(y) / f.tensor.tensors[mu-1].data
            else:
                b = []
        elif isinstance(self.training_data, list) and \
                len(self.training_data) == 2:
            b = self.training_data[1]
            self.linear_model_learning[mu-1].shared_coefficients = \
                mu != tree.root

        return self, A, b, f

    def set_parameter(self, f, mu, coef):
        f.tensor.tensors[mu-1] = \
            tensap.FullTensor(coef, shape=f.tensor.tensors[mu-1].shape)
        f.tensor.tensors[mu-1].is_orth = False
        return f

    def stagnation_criterion(self, f, f0):
        return (f.tensor - f0.tensor).norm() / f0.tensor.norm()

    def final_display(self, f):
        print('Ranks = [%s]' % ', '.join(map(str, f.tensor.ranks)), end='')

    def canonical_initialization(self, rank):
        '''
        Rank-r canonical initialization.

        Parameters
        ----------
        rank : int
            The rank of the canonical initialization.

        Returns
        -------
        tensap.TreeBasedTensor
            The rank-r canonical initialization..

        '''
        solver = tensap.CanonicalTensorLearning(self.order, self.loss_function)
        if isinstance(self.linear_model_learning, list):
            solver.linear_model_learning = self.linear_model_learning[0]
        else:
            solver.linear_model_learning = self.linear_model_learning
        solver.alternating_minimization_parameters = \
            deepcopy(self.alternating_minimization_parameters)
        solver.tolerance['on_stagnation'] = np.finfo(float).eps
        solver.tolerance['on_error'] = np.finfo(float).eps
        solver.bases = self.bases
        solver.bases_eval = self.bases_eval
        solver.bases_eval_test = self.bases_eval_test
        solver.display = False
        solver.alternating_minimization_parameters['display'] = False
        solver.initialization_type = 'mean'
        solver.rank_adaptation = True
        solver.rank_adaptation_options['max_iterations'] = rank
        solver.bases_adaptation_path = self.bases_adaptation_path
        solver.test_error = self.test_error
        solver.training_data = self.training_data
        solver.test_data = self.test_data
        solver._warnings = self._warnings

        f = solver.solve()[0]
        return f.tensor.tree_based_tensor(self.tree, self.is_active_node)

    def canonical_correction(self, f, rank):
        '''
        Rank-r canonical correction.

        Parameters
        ----------
        f : tensap.FunctionalTensor or None
            The current approximation.
        rank : int
            The rank of the canonical correction.

        Raises
        ------
        NotImplementedError
            If the method is not implemented.

        Returns
        -------
        f : tensap.FunctionalTensor
            The corrected approximation.

        '''
        if isinstance(f, tensap.FunctionalTensor):
            fx = f.tensor.tensor_matrix_product_eval_diag(self.bases_eval).data
        elif f is None:
            fx = 0
        else:
            raise NotImplementedError('Not implemented.')

        solver = deepcopy(self)
        if isinstance(solver.training_data, list) and \
                len(solver.training_data) == 2:
            solver.training_data[1] -= fx
        elif isinstance(solver.loss_function, tensap.DensityL2LossFunction):
            solver.training_data = [solver.training_data, f]

        f_add = solver.canonical_initialization(rank)
        if isinstance(f_add, tensap.FunctionalTensor):
            f_add = f_add.tensor
        if f is not None:
            f = f.tensor + f_add
        else:
            f = f_add
        return f

    def rank_one_correction(self, f):
        '''
        Rank one correction.

        Parameters
        ----------
        f : tensap.FunctionalTensor or None
            The current approximation.

        Raises
        ------
        NotImplementedError
            If the method is not implemented.

        Returns
        -------
        f : tensap.FunctionalTensor
            The corrected approximation.

        '''
        if isinstance(f, tensap.FunctionalTensor):
            fx = f.tensor.tensor_matrix_product_eval_diag(self.bases_eval).data
        elif f is None:
            fx = 0
        else:
            raise NotImplementedError('Not implemented.')

        solver = deepcopy(self)
        solver.model_selection = False
        if isinstance(solver.training_data, list) and \
                len(solver.training_data) == 2:
            solver.training_data[1] -= fx
        elif isinstance(solver.loss_function, tensap.DensityL2LossFunction):
            solver.training_data = [solver.training_data, f]
        solver.rank_adaptation = False
        solver.tree_adaptation = False
        solver.rank = 1
        solver.display = False
        solver.alternating_minimization_parameters['display'] = False
        solver.initialization_type = 'ones'
        solver.alternating_minimization_parameters['max_iterations'] = 1

        f_add = solver.solve()[0]
        if isinstance(f_add, tensap.FunctionalTensor):
            f_add = f_add.tensor
        if f is not None:
            f = f.tensor + f_add
        else:
            f = f_add
        return f

    def create_basis_adaptation_path(self, ranks, alpha):
        '''
        Creation of the basis adaptation path.

        Parameters
        ----------
        rank : list or numpy.ndarray
            The alpha-ranks of the current approximation.
        alpha : int
            The current node.

        Returns
        -------
        path : numpy.ndarray
            The basis adaptation path.

        '''
        tree = self.tree
        ranks = np.array(ranks)
        ranks[tree.root-1] = 1
        if tree.is_leaf[alpha-1]:
            path_alpha = self.bases_adaptation_path[
                np.nonzero(tree.dim2ind == alpha)[0][0]].astype(bool)
            ranks = ranks[alpha-1]
            path = path_alpha[np.newaxis, :, np.newaxis, :]
            path = np.tile(path, [1, 1, ranks, 2])
            path = np.reshape(path, [path.shape[1]*ranks, -1], order='F')
        else:
            assert not self.linear_model_learning_parameters[
                'basis_adaptation_internal_nodes'], \
                'Basis adaptation for internal nodes is not implemented.'
            ch = tree.children(alpha)
            if np.all(self.is_active_node[ch-1]):
                ch_a = ch[self.is_active_node[ch-1]]
                path = np.full(np.prod(ranks[ch_a-1]) * ranks[alpha-1], True)
            else:
                path_alpha = []
                for nod in ch:
                    if self.is_active_node[nod-1]:
                        path_alpha.append(np.full((ranks[nod-1], 1), True))
                    else:
                        ind = np.nonzero(nod == tree.dim2ind)[0][0]
                        path_alpha.append(
                            self.bases_adaptation_path[ind].astype(bool))
                path = path_alpha[-1]
                for ind in np.arange(len(path_alpha)-2, -1, -1):
                    path = np.kron(path, path_alpha[ind])
                path = np.tile(path, [ranks[alpha-1], 1]).astype(bool)
        return path

# %% Rank adaptation solver methods
    def local_solver(self):
        s_local = deepcopy(self)
        s_local.display = False
        s_local.rank_adaptation = False
        s_local.store_iterates = False
        s_local.test_error = False
        s_local.model_selection = False
        return s_local

    def new_rank_selection(self, f):
        if self.rank_adaptation_options['rank_one_correction']:
            s_local = deepcopy(self)
            ranks_add = np.ones(f.tensor.tree.nb_nodes, dtype=int)
            ranks_add[f.tensor.tree.root-1] = 0
            ranks_add[f.tensor.non_active_nodes-1] = 0
            s_local.rank = TreeBasedTensorLearning.make_ranks_admissible(
                f.tensor, f.tensor.ranks + ranks_add)[0]
            s_local.initialization_type = 'initial_guess'
            tr = tensap.Truncator(tolerance=0, max_rank=s_local.rank)
            s_local.initial_guess = tr.truncate(self.rank_one_correction(f))
            s_local.alternating_minimization_parameters['max_iterations'] = 10
            s_local.model_selection = False
            s_local.rank_adaptation = False
            s_local.display = False
            s_local.alternating_minimization_parameters['display'] = False
            tensor_for_selection = s_local.solve()[0].tensor
        else:
            tensor_for_selection = f.tensor

        sin_val = tensor_for_selection.singular_values()

        # Remove from the rank adaptation candidates: the inactive nodes, the
        # root, the leaf nodes with a rank equal to the dimension of the basis
        # associated to it, and the nodes for  which the smallest singular
        # value is almost zero.
        sin_val = np.array([np.nan if x is None else x for x in sin_val])
        sin_val[f.tensor.tree.root-1] = np.nan
        dim2ind = np.intersect1d(f.tensor.tree.dim2ind, f.tensor.active_nodes)
        ind = [len(set(x.size)) == 1 for x in f.tensor.tensors[dim2ind-1]]
        sin_val[dim2ind[ind]-1] = np.nan
        sin_val[s_local.rank != tensor_for_selection.ranks] = np.nan

        sin_val_min = np.array([np.min(x) for x in sin_val])
        sin_val_min[[x / tensor_for_selection.norm() < np.finfo(float).eps if
                     not np.isnan(x) else False for x in sin_val_min]] = np.nan

        # Remove nodes that cannot be enriched because their rank is equal to
        # the product of the ranks of their children, and their children cannot
        # be enriched themselves.
        tree = tensor_for_selection.tree
        rank = f.tensor.ranks
        desc = np.setdiff1d(np.arange(1, tree.nb_nodes+1),
                            np.nonzero(tree.is_leaf)[0]+1)
        cannot_be_increased = np.full(tree.nb_nodes, False)
        cannot_be_increased[tree.root-1] = True
        cannot_be_increased[tree.is_leaf] = np.isnan(sin_val_min[tree.is_leaf])
        for level in np.arange(np.max(tree.level), 0, -1):
            nod_lvl = np.intersect1d(tree.nodes_with_level(level), desc)
            for nod in nod_lvl:
                ch = tree.children(nod)
                if np.all(cannot_be_increased[ch-1]) and \
                        rank[nod-1] == np.prod(rank[ch-1]):
                    cannot_be_increased[nod-1] = True
        cannot_be_increased_nodes = tree.nodes_indices[cannot_be_increased]
        for level in np.arange(1, np.max(tree.level)):
            nod_lvl = np.intersect1d(tree.nodes_with_level(level),
                                     cannot_be_increased_nodes)
            for nod in nod_lvl:
                pa = tree.parent(nod)
                ind = np.setdiff1d(tree.children(pa), nod)
                ind = np.concatenate(([pa], ind))
                if np.all(cannot_be_increased[ind-1]) and \
                        rank[nod-1] == np.prod(rank[ind-1]):
                    cannot_be_increased[nod-1] = True
        sin_val_min[cannot_be_increased] = np.nan

        if np.all(np.isnan(sin_val_min)):
            enriched_nodes = np.array([], dtype=int)
        else:
            theta = self.rank_adaptation_options['theta'] * \
                np.nanmax(sin_val_min)
            enriched_nodes = np.nonzero([not np.isnan(x) and x >= theta for
                                         x in sin_val_min])[0] + 1

        new_rank = np.array(f.tensor.ranks)
        new_rank[enriched_nodes-1] += 1

        if not f.tensor.is_admissible_rank(new_rank):
            # Add to the already enriched nodes nodes one by one in decreasing
            # order of singular value until the rank is admissible.
            enriched_nodes_theta = np.array(enriched_nodes)
            rank_theta = np.array(new_rank)
            sin_val_min[enriched_nodes_theta-1] = np.nan
            sin_val_min_sorted = TreeBasedTensorLearning.unique_tol(
                sin_val_min, 1e-2)
            sin_val_min_sorted = np.flip(sin_val_min_sorted)
            sin_val_min_sorted = sin_val_min_sorted[
                np.isfinite(sin_val_min_sorted)]

            for sv in sin_val_min_sorted:
                new_rank = np.array(rank_theta)
                ind = [x >= sv if not np.isnan(x) else False for
                       x in sin_val_min]
                new_rank[ind] += 1
                if f.tensor.is_admissible_rank(new_rank):
                    enriched_nodes = np.concatenate((enriched_nodes_theta,
                                                     np.nonzero(ind)[0]+1))
                    break
            if not f.tensor.is_admissible_rank(new_rank):
                new_rank = f.tensor.ranks
                enriched_nodes = np.array([])

        return new_rank, enriched_nodes, tensor_for_selection

    def initial_guess_new_rank(self, s_local, f, new_rank):
        s_local.initialization_type = 'initial_guess'
        if not np.all(f.ranks == new_rank):
            tr = tensap.Truncator(tolerance=0, max_rank=new_rank)
            s_local.initial_guess = tr.truncate(f)
        else:
            s_local.initial_guess = f
        return s_local

    def adaptation_display(self, f, enriched_nodes):
        print('\tEnriched nodes: [%s]\n\tRanks = [%s]' %
              (', '.join(map(str, enriched_nodes)),
               ', '.join(map(str, f.tensor.ranks))), flush=True)

    def adapt_tree(self, f, cv_error, test_error, output, *args):
        if not self.tree_adaptation:
            return self, f, output

        output['adapted_tree'] = False

        if np.any(f.tensor.ranks[f.tensor.active_nodes-1] == 0):
            print('Some ranks equal to 0, disabling tree adaptation for ' +
                  'this step.')
            return self, f, output
        if self.tree_adaptation_options['tolerance'] is not None:
            adapt_tree_error = self.tree_adaptation_options['tolerance']
        elif self.loss_function.error_type == 'relative':
            if test_error is None or test_error == 0:
                adapt_tree_error = cv_error
            elif cv_error is None or test_error != 0:
                adapt_tree_error = test_error
        else:
            print('Must provide a tolerance for the tree adaptation in ' +
                  'the treeAdaptationOptions property. Disabling tree ' +
                  'adaptation.')
            self.tree_adaptation = False
            return self, f, output

        f_perm = f.tensor.optimize_dimension_tree(
            adapt_tree_error, self.tree_adaptation_options['max_iterations'])
        if f_perm.storage() < f.tensor.storage():
            f.tensor = f_perm
            self.tree = f.tensor.tree
            self.is_active_node = f.tensor.is_active_node
            output['adapted_tree'] = True
            if self.display:
                print('\tTree adaptation:\n\t\tRanks after permutation ' +
                      '= [%s]' % ', '.join(map(str, f.tensor.ranks)))
        return self, f, output

# %% Inner rank adaptation solver
    def _solve_dmrg_rank_adaptation(self):
        if 'max_rank' not in self.rank_adaptation_options:
            self.rank_adaptation_options['max_rank'] = 100
        if 'post_alternating_minimization' not in self.rank_adaptation_options:
            self.rank_adaptation_options['post_alternating_minimization'] = \
                False
        if self.rank_adaptation_options['type'] == 'dmrg_low_rank' and \
                'model_selection_type' not in self.rank_adaptation_options:
            self.rank_adaptation_options['model_selection_type'] = 'cv_error'

        if self.display:
            self.alternating_minimization_parameters['display'] = True

        output = {'flag': 0}

        self, f = self.initialize()
        f = tensap.FunctionalTensor(f, self.bases_eval)

        # Exploration strategy of the tree by decreasing level
        tree = f.tensor.tree
        exploration_strategy = np.zeros(self._number_of_parameters, dtype=int)
        active_nodes = f.tensor.active_nodes
        rep = 0
        for level in np.arange(np.max(tree.level), -1, -1):
            nodes = np.intersect1d(tree.nodes_with_level(level), active_nodes)
            exploration_strategy[rep:rep+len(nodes)] = nodes
            rep += len(nodes)
        self._exploration_strategy = np.setdiff1d(exploration_strategy,
                                                  tree.root,
                                                  assume_unique=True)

        # Replication of the LinearModelLearning objects
        if self.linear_model_learning_parameters[
                'identical_for_all_parameters'] and \
                not isinstance(self.linear_model_learning, (list, np.ndarray)):
            self.linear_model_learning = list(map(deepcopy, [
                self.linear_model_learning] * self._number_of_parameters))
        elif isinstance(self.linear_model_learning, (list, np.ndarray)) and \
                len(self.linear_model_learning) != self._number_of_parameters:
            raise ValueError('Must provide self._numberOfParameters ' +
                             'LinearModelLearning objects.')

        if self.error_estimation:
            for x in self.linear_model_learning:
                setattr(x, 'error_estimation', True)

        # Working set paths
        if isinstance(self.linear_model_learning, (list, np.ndarray)) and \
            np.any([x.basis_adaptation for
                    x in self.linear_model_learning]) and \
                self.bases_adaptation_path is None:
            self.bases_adaptation_path = self.bases.adaptation_path()

        if self.alternating_minimization_parameters['max_iterations'] == 0:
            return f, output

        output['stagnation_criterion'] = []
        output['iterates'] = []
        output['error_iterations'] = []
        output['test_error_iterations'] = []

        # Alternating minimization loop
        for iteration in range(self.alternating_minimization_parameters
                               ['max_iterations']):
            self, f = self.pre_processing(f)
            f0 = deepcopy(f)

            if self.alternating_minimization_parameters['random']:
                # Randomize the exploration strategy
                alpha_list = self.randomize_exploration_strategy()
            else:
                alpha_list = self._exploration_strategy

            tree = f.tensor.tree
            for alpha in alpha_list:
                if self.linear_model_learning[alpha-1].basis_adaptation:
                    if np.isin(alpha, tree.internal_nodes):
                        if self.linear_model_learning_parameters[
                                'basis_adaptation_internal_nodes']:
                            tr = tensap.Truncator(
                                tolerance=np.finfo(float).eps,
                                max_rank=np.max(f.tensor.ranks))
                            f.tensor = tr.hsvd(f.tensor)
                        elif np.all(f.tensor.is_active_node[
                                tree.children(alpha)-1]):
                            self.linear_model_learning[alpha-1].\
                                basis_adaptation = False
                    f.tensor = f.tensor.orth_at_node(tree.parent(alpha))
                    self.tree = tree
                    self.is_active_node = f.tensor.is_active_node

                    if self.rank_adaptation_options['type'] == 'dmrg':
                        self.linear_model_learning[alpha-1].\
                            basis_adaptation_path = \
                            self.create_basis_adaptation_path_dmrg(
                                f.tensor.ranks, alpha)
                    elif self.rank_adaptation_options['type'] == \
                            'dmrg_low_rank':
                        self.linear_model_learning[alpha-1].\
                            basis_adaptation_path = self.\
                            create_basis_adaptation_path_dmrg_low_rank(
                                f.tensor.ranks, alpha)
                    else:
                        raise ValueError('Wrong rank adaptation type.')
                else:
                    f.tensor = f.tensor.orth_at_node(tree.parent(alpha))

                grad = f.parameter_gradient_eval_dmrg(
                    alpha, dmrg_type=self.rank_adaptation_options['type'])

                if isinstance(self.loss_function,
                              tensap.DensityL2LossFunction):
                    if isinstance(self.training_data, list) and \
                            len(self.training_data) == 2:
                        y = self.training_data[1]
                        if isinstance(y, tensap.FunctionalTensor):
                            y = y.tensor
                        y = y.orth()
                        if tree.is_leaf[alpha-1]:
                            a = deepcopy(y)
                            for nod in tree.internal_nodes:
                                a.tensors[nod-1] = tensap.FullTensor(
                                    y.tensors[nod-1].data *
                                    f.tensor.tensors[nod-1].data,
                                    shape=y.tensors[nod-1].shape)
                            ind = np.setdiff1d(
                                range(self.order),
                                np.nonzero(tree.dim2ind == alpha)[0])
                            C = [x.data for
                                 x in f.tensor.tensors[tree.dim2ind-1]]
                            b = a.tensor_vector_product([C[x] for x in ind],
                                                        ind)
                            b = b.tensors[0].data
                        else:
                            b = f.tensor.dot(y) / \
                                f.tensor.tensors[alpha-1].data
                    else:
                        b = []
                elif isinstance(self.training_data, list) and \
                        len(self.training_data) == 2:
                    b = self.training_data[1]
                    self.linear_model_learning[alpha-1].shared_coefficients = \
                        alpha != tree.root

                gamma = tree.parent(alpha)
                ind = np.setdiff1d(
                    np.arange(1, f.tensor.tensors[gamma-1].order+1),
                    tree.child_number(alpha))

                if self.rank_adaptation_options['type'] == 'dmrg':
                    A = np.reshape(grad.data, (grad.shape[0], -1), order='F')

                    self.linear_model_learning[alpha-1].training_data = [None,
                                                                         b]
                    self.linear_model_learning[alpha-1].basis = None
                    self.linear_model_learning[alpha-1].basis_eval = A

                    C, output_tmp = \
                        self.linear_model_learning[alpha-1].solve()

                    if C is None or np.count_nonzero(C) == 0 or \
                            not np.all(np.isfinite(C)):
                        print('Empty, zero or NaN solution, returning to ' +
                              'the previous iteration.')
                        output['flag'] = -2
                        output['error'] = np.inf
                        break

                    sz_1 = np.prod(f.tensor.tensors[alpha-1].shape[:-1])
                    sz_2 = np.prod(f.tensor.tensors[gamma-1].shape[ind-1])
                    C = tensap.FullTensor(C, 2, [sz_1, sz_2])

                    tr = tensap.Truncator()
                    tr.tolerance = self.tolerance['on_error'] / \
                        np.sqrt(np.count_nonzero(f.tensor.is_active_node)-1)
                    tr.max_rank = self.rank_adaptation_options['max_rank']
                    C = tr.truncate(C)
                    rank = [np.shape(C.space[0])[1]]

                    sz_1 = np.concatenate(
                        (f.tensor.tensors[alpha-1].shape[:-1], rank))
                    sz_2 = np.concatenate(
                        (f.tensor.tensors[gamma-1].shape[ind-1], rank))

                    a_alpha = np.reshape(C.space[0], sz_1, order='F')
                    a_gamma = np.reshape(C.space[1]*np.tile(
                        np.reshape(C.core.data, [1, -1]),
                        (np.shape(C.space[1])[0], 1)), sz_2, order='F')
                    perm = np.concatenate((ind, [tree.child_number(alpha)]))
                    a_gamma = np.transpose(a_gamma, np.argsort(perm-1))
                    sz_2[perm-1] = np.array(sz_2)
                elif self.rank_adaptation_options['type'] == 'dmrg_low_rank':
                    A = [np.reshape(x.data, [x.shape[0], -1], order='F') for
                         x in grad]

                    s_local = TreeBasedTensorLearning(
                        tensap.DimensionTree.trivial(2), [True, True, False],
                        self.loss_function)
                    s_local.rank_adaptation = True
                    s_local.tolerance['on_error'] = \
                        self.tolerance['on_error'] / \
                        np.sqrt(np.count_nonzero(f.tensor.is_active_node)-1)
                    s_local.tolerance['on_stagnation'] = \
                        self.tolerance['on_stagnation']
                    s_local.rank_adaptation_options['max_iterations'] = \
                        self.rank_adaptation_options['max_rank']
                    s_local.alternating_minimization_parameters = \
                        deepcopy(self.alternating_minimization_parameters)
                    s_local.alternating_minimization_parameters['display'] = \
                        False
                    s_local.store_iterates = True
                    s_local.test_error = False
                    s_local.error_estimation = True
                    s_local.display = False
                    s_local.order = 2
                    s_local.linear_model_learning = \
                        self.linear_model_learning[alpha-1]
                    s_local._warnings['orthonormality_warning_display'] = False
                    s_local._warnings['empty_bases_warning_display'] = False
                    s_local.bases_adaptation_path = \
                        self.linear_model_learning[alpha-1].\
                        basis_adaptation_path
                    s_local.training_data = self.training_data
                    s_local.bases_eval = A
                    s_local.model_selection = True
                    s_local.model_selection_options['type'] = \
                        self.rank_adaptation_options['model_selection_type']
                    C, output_tmp = s_local.solve()

                    rank = [C.tensor.ranks[1]]
                    sz_1 = np.concatenate(
                        (f.tensor.tensors[alpha-1].shape[:-1], rank))
                    sz_2 = np.concatenate(
                        (f.tensor.tensors[gamma-1].shape[ind-1], rank))

                    a_alpha = C.tensor.tensors[1].reshape(sz_1)
                    a_gamma = C.tensor.tensors[0].transpose([1, 0]).reshape(
                        sz_2)

                    perm = np.concatenate((ind, [tree.child_number(alpha)]))
                    a_gamma = a_gamma.itranspose(perm-1)
                    sz_2[perm-1] = np.array(sz_2)

                    a_alpha = a_alpha.data
                    a_gamma = a_gamma.data

                    if (a_alpha is None or np.count_nonzero(a_alpha) == 0 or
                            not np.all(np.isfinite(a_alpha))) or \
                        (a_gamma is None or np.count_nonzero(a_gamma) == 0 or
                            not np.all(np.isfinite(a_gamma))):
                        print('Empty, zero or NaN solution, returning to ' +
                              'the previous iteration.')
                        output['flag'] = -2
                        output['error'] = np.inf
                        break
                else:
                    raise ValueError('Wrong rank adaptation type.')
                f.tensor.tensors[alpha-1] = tensap.FullTensor(a_alpha,
                                                              np.size(sz_1),
                                                              sz_1)
                f.tensor.tensors[gamma-1] = tensap.FullTensor(a_gamma,
                                                              np.size(sz_2),
                                                              sz_2)

            if self.rank_adaptation_options['post_alternating_minimization']:
                print('\t\tPost alternating minimization.')
                s_local = TreeBasedTensorLearning(self.tree,
                                                  self.is_active_node,
                                                  self.loss_function)
                s_local.bases_eval = self.bases_eval
                s_local.bases_adaptation_path = self.bases_adaptation_path
                s_local.training_data = self.training_data
                s_local.test_error = self.test_error
                s_local.test_data = self.test_data
                s_local.bases_eval_test = self.bases_eval_test
                s_local.model_selection = False
                s_local.rank_adaptation = False
                s_local.store_iterates = False
                s_local.rank = f.tensor.ranks
                s_local.initialization_type = 'initial_guess'
                s_local.initial_guess = f.tensor
                s_local.tolerance['on_error'] = self.tolerance['on_error'] / \
                    np.sqrt(np.count_nonzero(f.tensor.is_active_node)-1)
                s_local.tolerance['on_stagnation'] = \
                    self.tolerance['on_stagnation']
                s_local.alternating_minimization_parameters = \
                    deepcopy(self.alternating_minimization_parameters)
                s_local.alternating_minimization_parameters['display'] = False
                s_local.error_estimation = True
                s_local.display = False
                s_local.linear_model_learning = \
                    [x for x in self.linear_model_learning if x is not None]
                s_local._warnings['orthonormality_warning_display'] = False
                s_local._warnings['empty_bases_warning_display'] = False
                f, output_tmp = s_local.solve()

            stagnation = self.stagnation_criterion(f, f0)
            output['stagnation_criterion'].append(stagnation)

            if self.store_iterates:
                if isinstance(self.bases, tensap.FunctionalBases):
                    output['iterates'].append(tensap.FunctionalTensor(
                        f.tensor, self.bases))
                else:
                    output['iterates'].append(f)

            if 'error' in output_tmp:
                output['error'] = output_tmp['error']
                output['error_iterations'].append(output['error'])

            if self.test_error:
                f_eval_test = tensap.FunctionalTensor(f, self.bases_eval_test)
                output['test_error'] = self.loss_function.test_error(
                    f_eval_test, self.test_data)
                output['test_error_iterations'].append(output['test_error'])

            if self.alternating_minimization_parameters['display']:
                print('\tAlt. min. iteration %s/%i: stagnation = %2.5e' %
                      (str(iteration).
                       zfill(len(str(self.alternating_minimization_parameters
                                     ['max_iterations']-1))),
                       self.alternating_minimization_parameters
                       ['max_iterations']-1,
                       stagnation), end='')
                if 'error' in output:
                    print(', error = %2.5e' % output['error'], end='')
                if self.test_error:
                    if not np.isscalar(output['test_error']):
                        output['test_error'] = output['test_error'].numpy()
                    print(', test error = %2.5e' % output['test_error'],
                          end='')
                print('')

            if iteration > 0 and stagnation < \
                    self.alternating_minimization_parameters['stagnation']:
                output['flag'] = 1
                break

            if self.test_error and \
                    output['test_error'] < self.tolerance['on_error']:
                output['flag'] = 1
                break

            if self.tree_adaptation and iteration > 0:
                C_old = f.tensor.storage()
                self, f, output = self.adapt_tree(f, output['error'],
                                                  None, output, iteration)
                adapted_tree = output['adapted_tree']
                if adapted_tree:
                    if self.display:
                        print('\t\tStorage complexity before permutation ' +
                              '= %i' % C_old)
                        print('\t\tStorage complexity after permutation ' +
                              '= %i' % f.tensor.storage())
                    if self.test_error:
                        f_eval_test = tensap.FunctionalTensor(
                            f, self.bases_eval_test)
                        if self.display:
                            print('\t\tTest error after permutation ' +
                                  '= %2.5e' % self.loss_function.test_error(
                                      f_eval_test, self.test_data))

                    if self.alternating_minimization_parameters['display']:
                        print('')

        if isinstance(self.bases, tensap.FunctionalBases):
            f = tensap.FunctionalTensor(f.tensor, self.bases)
        output['iter'] = iteration

        if 'adapted_tree' in output:
            del output['adapted_tree']

        if self.display and not self.model_selection:
            if self.alternating_minimization_parameters['display']:
                print('')
            self.final_display(f)
            if 'error' in output:
                print(', error = %2.5e' % output['error'], end='')
            if 'test_error' in output:
                print(', test error = %2.5e' % output['test_error'], end='')
            print('')

        return f, output

    def create_basis_adaptation_path_dmrg(self, ranks, alpha):
        tree = self.tree
        ranks = np.array(ranks)
        ranks[tree.root-1] = 1
        gamma = tree.parent(alpha)

        p_alpha = self.create_basis_adaptation_path(ranks, alpha)
        if p_alpha.ndim == 1:
            p_alpha = np.reshape(p_alpha, [-1, 1])
        p_alpha = p_alpha[:int(p_alpha.shape[0]/ranks[alpha-1]), :]

        ch = tree.children(gamma)
        p_gamma = [None]*len(ch)
        for nod in np.setdiff1d(ch, alpha):
            ind = tree.child_number(nod) < tree.child_number(alpha)
            if self.is_active_node[nod-1]:
                p_gamma[tree.child_number(nod)-1+ind] = \
                    np.full((ranks[nod-1], 1), True)
            else:
                rep = np.nonzero(nod == tree.dim2ind)[0][0]
                p_gamma[tree.child_number(nod)-1+ind] = \
                    self.bases_adaptation_path[rep].astype(bool)
        p_gamma[0] = p_alpha

        r_gamma = ranks[gamma-1]
        p = p_gamma[-1]
        for i in np.arange(len(p_gamma)-2, -1, -1):
            p = np.kron(p, p_gamma[i])
        return np.tile(p, (r_gamma, 1)).astype(bool)

    def create_basis_adaptation_path_dmrg_low_rank(self, ranks, alpha):
        tree = self.tree
        ranks = np.array(ranks)
        ranks[tree.root-1] = 1
        gamma = tree.parent(alpha)

        p_alpha = self.create_basis_adaptation_path(ranks, alpha)
        if p_alpha.ndim == 1:
            p_alpha = np.reshape(p_alpha, [-1, 1])
        p_alpha = p_alpha[:int(p_alpha.shape[0]/ranks[alpha-1]), :]

        ch = np.setdiff1d(tree.children(gamma), alpha)
        p_gamma = [None]*len(ch)
        for nod in ch:
            ind = tree.child_number(nod) > tree.child_number(alpha)
            if self.is_active_node[nod-1]:
                p_gamma[tree.child_number(nod)-1-ind] = \
                    np.full((ranks[nod-1], 1), True)
            else:
                rep = np.nonzero(nod == tree.dim2ind)[0][0]
                p_gamma[tree.child_number(nod)-1-ind] = \
                    self.bases_adaptation_path[rep].astype(bool)

        r_gamma = ranks[gamma-1]
        p = p_gamma[-1]
        for i in np.arange(len(p_gamma)-2, -1, -1):
            p = np.kron(p, p_gamma[i])
        p = np.tile(p, (r_gamma, 1)).astype(bool)

        return [p_alpha, p]

# %% Static methods
    @staticmethod
    def tensor_train(order, *args):
        '''
        Call of the constructor of the class TreeBasedTensorLearning, with a
        tree and active nodes corresponding to the Tensor-Train format in
        dimension order.

        See also TreeBasedTensorLearning.

        Parameters
        ----------
        order : int
            The order of the tensor.
        *args : tuple
            Additional parameters (see the constructor of
            TreeBasedTensorLearning).

        Returns
        -------
        TreeBasedTensorLearning
            The solver with a tree and active nodes associated with the
            Tensor-Train format.

        '''
        tree = tensap.DimensionTree.linear(order)
        is_active_node = np.full(tree.nb_nodes, True)
        is_active_node[tree.dim2ind[1:]-1] = False
        return TreeBasedTensorLearning(tree, is_active_node, *args)

    @staticmethod
    def tensor_train_tucker(order, *args):
        '''
        Call of the constructor of the class TreeBasedTensorLearning, with a
        tree and active nodes corresponding to the Tensor-Train Tucker format
        in dimension order.

        See also TreeBasedTensorLearning.

        Parameters
        ----------
        order : int
            The order of the tensor.
        *args : tuple
            Additional parameters (see the constructor of
            TreeBasedTensorLearning).

        Returns
        -------
        TreeBasedTensorLearning
            The solver with a tree and active nodes associated with the
            Tensor-Train Tucker format.

        '''
        tree = tensap.DimensionTree.linear(order)
        is_active_node = np.full(tree.nb_nodes, True)
        return TreeBasedTensorLearning(tree, is_active_node, *args)

    @staticmethod
    def enriched_edges_to_ranks_random(f, new_rank):
        '''
        Enrichment of the ranks of specified edges of the tensor f using random
        additions for each child / parent couple of the enriched edges.

        Parameters
        ----------
        f : tensap.TreeBasedTensor
            The tree-based tensor to enrich.
        new_rank : list or numpy.ndarray
            The new tree-based rank.

        Returns
        -------
        f : tensap.TreeBasedTensor
            The enriched tree-based tensor.

        '''
        f.is_orth = False
        tree = f.tree
        enriched_dims = np.nonzero(new_rank > f.ranks)[0]

        for level in np.arange(1, np.max(tree.level)+1):
            nod_lvl = np.intersect1d(tree.nodes_with_level(level),
                                     enriched_dims)
            for alpha in nod_lvl:
                gamma = tree.parent(alpha)
                rank = new_rank[alpha-1] - f.ranks[alpha-1]

                A = np.reshape(f.tensors[alpha-1].data,
                               [-1, f.ranks[alpha-1]], order='F')
                A = np.hstack((A, np.tile(np.reshape(A[:, -1], [-1, 1]),
                                          [1, rank]) *
                               (1+np.random.randn(A.shape[0], rank))))
                A[:, -1-rank-1:] /= np.sqrt(np.sum(A[:, -1-rank-1:]**2, 0))
                shape = np.array(f.tensors[alpha-1].shape)
                shape[-1] += rank
                f.tensors[alpha-1].data = np.reshape(A, shape, order='F')

                ch = f.tree.child_number(alpha)-1
                ind = np.setdiff1d(range(f.tensors[gamma-1].order), ch)
                ind = np.concatenate((ind, [ch]))
                A = np.transpose(f.tensors[gamma-1].matricize(ch).data)
                A = np.hstack((A, np.tile(np.reshape(A[:, -1], [-1, 1]),
                                          [1, rank]) *
                               (1+np.random.randn(A.shape[0], rank))))
                A[:, -1-rank-1:] /= np.sqrt(np.sum(A[:, -1-rank-1:]**2, 0))
                shape = np.array(f.tensors[gamma-1].shape)
                shape[ch] += rank
                A = np.reshape(A, shape[ind], order='F')
                f.tensors[gamma-1].data = np.transpose(A, np.argsort(ind))

                f = f.update_attributes()
        return f

    @staticmethod
    def make_ranks_admissible(f, rank):
        '''
        Adjustment of the ranks to make the associated tree-based tensor f
        rank-admissible, by enriching new edges associated with nodes of the
        tree until all the rank admissibility conditions are met.

        Parameters
        ----------
        f : tensap.TreeBasedTensor
            The current approximation in tree-based tensor format.
        rank : numpy.ndarray
            The proposed tree-based rank.

        Returns
        -------
        rank : numpy.ndarray
            The admissible tree-based rank (if possible).
        d : numpy.ndarray
            The enriched nodes.

        '''
        # Do not increase the ranks of leaf nodes with rank equal to the
        # dimension of the approximation space.
        nodes = f.active_nodes
        ind = [f.tree.is_leaf[x] and rank[x] > f.tensors[x].shape[0] for
               x in nodes-1]
        rank[nodes[ind]-1] = [x.shape[0] for x in f.tensors[nodes[ind]-1]]
        rank[np.logical_not(f.is_active_node)] = 0

        delta = rank - f.ranks
        if f.is_admissible_rank(f.ranks + delta):
            rank = f.ranks + delta
            d = np.nonzero(delta)[0] + 1
            return rank, d

        ind = np.nonzero(delta)[0]
        for i in np.arange(1, np.count_nonzero(delta)+1):
            pos = list(combinations(np.arange(np.count_nonzero(delta)), i))
            shuffle(pos)
            for pos_loc in pos:
                delta_loc = np.array(delta)
                delta_loc[ind[list(pos_loc)]] = 0
                if f.is_admissible_rank(f.ranks + delta_loc):
                    rank = f.ranks + delta_loc
                    d = np.nonzero(delta_loc)[0] + 1
                    return rank, d
        return rank, []

    @staticmethod
    def unique_tol(inp, tol):
        '''
        Unique values within tolerance, with sorted output.

        Parameters
        ----------
        inp : list of numpy.ndarray
            The values to be checked.
        tol : float
            The tolerance defining the uniqueness.

        Returns
        -------
        bout : numpy.ndarray
            The unique values of inp, within the tolerance tol.

        '''
        inp = np.sort(inp)
        out = [inp[0]]

        for ind in np.arange(1, inp.size):
            if np.abs(out[-1] - inp[ind]) / out[-1] > tol:
                out = np.concatenate((out, [inp[ind]]))
        return out



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module linear_model_learning_density_l2.

'''

import numpy as np
import tensap


class LinearModelLearningDensityL2(tensap.LinearModelLearning):
    '''
    Class LinearModelLearningDensityL2.

    Attributes
    ----------
    is_basis_orthonormal : bool, optional
        Boolean indicating if the basis used to compute the approximation
        is orthonormal according to some reference measure. The default is
        True.

    '''

    def __init__(self):
        '''
        Constructor for the class LinearModelLearningDensityL2.

        Returns
        -------
        None.

        '''
        super().__init__(tensap.SquareLossFunction())

        self.is_basis_orthonormal = True

    def solve(self):
        '''
        Solution (Ordinary or Regularized) of the minimization problem and
        cross-validation procedure.

        Returns
        -------
        sol : numpy.ndarray  or tensap.FunctionalBasisArray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        self.initialize()

        if self.basis_adaptation:
            sol, output = self._solve_basis_adaptation()
        elif self.regularization:
            sol, output = self._solve_regularized()
        else:
            sol, output = self._solve_standard()

        if self.test_error:
            f_eval = np.matmul(self.basis_eval_test, sol)
            output['test_error'] = self.loss_function.test_error(
                f_eval, self.test_data, sol.norm()**2)

        if self.basis is not None:
            sol = tensap.FunctionalBasisArray(sol, self.basis)

        return sol, output

    def _solve_standard(self):
        '''
        Solution of the minimization problem and cross-validation procedure
        (if requested).

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        assert self.is_basis_orthonormal, \
            'Only implemented for orthonormal bases.'

        A = self.basis_eval
        if np.ndim(A) == 3:
            A = np.squeeze(A, axis=2)

        sol = np.mean(A, axis=0)
        if isinstance(self.training_data, list) and \
            len(self.training_data) == 2 and \
                np.size(self.training_data[1]):
            b = np.reshape(self.training_data[1], sol.shape)

            sol -= b
        else:
            b = []

        output = {}
        if self.error_estimation and self.error_estimation_type == 'leave_out':
            N = A.shape[0]
            if np.size(b) == 0:
                output['error'] = (-N**2)/(1-N)**2*np.linalg.norm(sol)**2 + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(np.ravel(A)**2)
            else:
                output['error'] = (N**2-2*N)/(N-1)**2*np.linalg.norm(sol)**2 +\
                    1/(N-1)**2*np.linalg.norm(b)**2 - \
                    2/(N-1)**2*np.sum(np.matmul(A, b)) + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(np.ravel(A)**2) - \
                    2/(N-1)*np.sum(np.matmul(A, sol)) + \
                    2*np.sum(sol*b)

        return sol, output

    def _solve_regularized(self):
        '''
        Solution of the regularized minimization problem and cross-validation
        procedure.

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        sol_standard, _ = self._solve_standard()

        A = self.basis_eval
        if np.ndim(A) == 3:
            A = np.squeeze(A, axis=2)
        N = A.shape[0]

        if isinstance(self.training_data, list) and \
            len(self.training_data) == 2 and \
                np.size(self.training_data[1]):
            b = np.reshape(self.training_data[1], sol_standard.shape)
        else:
            b = []

        A_square_sum = np.sum(A**2, 0)

        list_sort = np.argsort(-np.abs(sol_standard))

        if 'included_coefficients' in self.regularization_options:
            incl_coef = self.regularization_options['included_coefficients']
            list_sort = list_sort[np.logical_not(np.in1d(list_sort,
                                                         incl_coef))]

            sol_incl_coef = np.array(sol_standard)
            sol_incl_coef[np.setdiff1d(range(len(sol_incl_coef)),
                                       incl_coef)] = 0

            if np.size(b) == 0:
                err_incl_coef = \
                    (-N**2)/(1-N)**2*np.linalg.norm(sol_incl_coef)**2 + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(
                        np.ravel(A_square_sum)[incl_coef])
            else:
                b_incl_coef = b[incl_coef]
                err_incl_coef = \
                    (N**2-2*N)/(N-1)**2*np.linalg.norm(sol_incl_coef)**2 +\
                    1/(N-1)**2*np.linalg.norm(b_incl_coef)**2 - \
                    2/(N-1)**2*np.sum(
                        np.matmul(A[:, incl_coef], b_incl_coef)) + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(
                        np.ravel(A_square_sum)[incl_coef]) -\
                    2/(N-1)*np.sum(np.matmul(A, sol_incl_coef)) + \
                    2*np.sum(sol_incl_coef*b_incl_coef)

        if list_sort.size == 0:
            return sol_standard, {'error': np.nan}

        n = list_sort.size
        sol = np.zeros((sol_standard.shape[0], n))
        err = np.zeros(n)

        for i in range(n):
            ind = list_sort[:i+1]
            if 'included_coefficients' in self.regularization_options:
                ind = np.concatenate((incl_coef, ind))
            sol_red = sol_standard[ind]

            if np.size(b) == 0:
                err[i] = \
                    (-N**2)/(1-N)**2*np.linalg.norm(sol_red)**2 + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(
                        np.ravel(A_square_sum)[ind])
            else:
                err[i] = \
                    (N**2-2*N)/(N-1)**2*np.linalg.norm(sol_red)**2 +\
                    1/(N-1)**2*np.linalg.norm(b[ind])**2 - \
                    2/(N-1)**2*np.sum(np.matmul(A[:, ind], b[ind])) + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(np.ravel(A_square_sum)[ind]) -\
                    2/(N-1)*np.sum(np.matmul(A[:, ind], sol_red)) + \
                    2*np.sum(sol_red*b[ind])

            sol[ind, i] = np.squeeze(sol_red)

        if 'included_coefficients' in self.regularization_options:
            sol = np.hstack((np.reshape(sol_incl_coef, (-1, 1)), sol))
            err = np.hstack((err_incl_coef, err))

        ind = np.argmin(err)
        pattern = sol != 0

        output = {}
        output['error'] = err[ind]
        output['error_path'] = err
        output['ind'] = ind
        output['pattern'] = pattern[:, ind]
        output['pattern_path'] = pattern
        output['optimal_solution'] = sol[:, ind]
        output['solution_path'] = sol

        return sol[:, ind], output

    def _solve_basis_adaptation(self):
        '''
        Solution of the minimization problem with working-set and
        cross-validation procedure.

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        if self.basis_adaptation_path is None:
            solpath = np.triu(np.full([self.basis_eval.shape[1]]*2, True))
        else:
            solpath = self.basis_adaptation_path

        sol, output = self._select_optimal_path(solpath)
        output['flag'] = 2
        return sol, output

    def _select_optimal_path(self, solpath):
        '''
        Selection of a solution using leave-one-out cross-validation error.

        The dictionnary output contains:
            - error: the leave-one-out cross-validation error estimate,
            - error_path: the regularization path of leave-one-out
            cross-validation error estimates,
            - ind: the index of the optimal pattern,
            - pattern: the optimal sparsity pattern,
            - pattern_path: the regularization path of sparsity patterns,
            - solution_path: the regularization path of the solutions,
            - optimal_solution: the optimal solution.

        Parameters
        ----------
        solpath : numpy.ndarray
            Array of shape (P, m) whose columns give m potential solutions with
            different sparsity patterns.

        Returns
        -------
        numpy.ndarray
            The optimal solution of the minimization problem.
        dict
            Outputs of the algorithm.

        '''
        sol_standard, output = self._solve_standard()

        A = self.basis_eval
        if np.ndim(A) == 3:
            A = np.squeeze(A, axis=2)

        if isinstance(self.training_data, list) and \
            len(self.training_data) == 2 and \
                np.size(self.training_data[1]):
            b = np.reshape(self.training_data[1], sol_standard.shape)
        else:
            b = []

        A_square_sum = np.sum(A**2, 0)
        if np.size(b) != 0:
            A_sum = np.sum(A, 0)

        pattern = solpath != 0
        rep = np.any(pattern, 0)
        pattern = pattern[:, rep]

        _, ind = np.unique(pattern, return_index=True, axis=1)
        ind = np.sort(ind)
        pattern = pattern[:, ind]

        ind = np.sum(pattern != 0, 0) <= A.shape[0]
        pattern = pattern[:, ind]

        if np.linalg.norm(pattern) == 0:
            return sol_standard, output

        sol = np.zeros(pattern.shape)
        err = np.zeros(pattern.shape[1])

        N = A.shape[0]

        for i in range(pattern.shape[1]):
            ind = pattern[:, i]
            sol_red = sol_standard[ind]

            if np.size(b) == 0:
                err[i] = (-N**2)/(1-N)**2*np.linalg.norm(sol_red)**2 + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(np.ravel(A_square_sum)[ind])
            else:
                b_red = b[ind]
                err[i] = (N**2-2*N)/(N-1)**2*np.linalg.norm(sol_red)**2 +\
                    1/(N-1)**2*np.linalg.norm(b_red)**2 - \
                    2/(N-1)**2*np.sum(A_sum[ind]*b_red) + \
                    (2*N-1)/(N*(N-1)**2)*np.sum(np.ravel(A_square_sum)[ind]) -\
                    2/(N-1)*np.sum(A_sum[ind]*sol_red) + \
                    2*np.sum(sol_red*b_red)
            sol[ind, i] = np.squeeze(sol_red)

            if i > 1 and \
                self.model_selection_options['stop_if_error_increase'] and \
                    err[i] > 2*err[i-1]:
                print('stop_if_error_increase')
                err[i+1:] = np.inf

        ind = np.argmin(err)

        output['error'] = err[ind]
        output['error_path'] = err
        output['ind'] = ind
        output['pattern'] = pattern[:, ind]
        output['pattern_path'] = pattern
        output['optimal_solution'] = sol[:, ind]
        output['solution_path'] = sol

        return sol[:, ind], output



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module polynomials.

'''

from abc import ABC, abstractmethod
import numpy as np
import tensap


class UnivariatePolynomials(ABC):
    '''
    Class UnivariatePolynomials.

    '''

    def polyval(self, ind, x):
        '''
        Evaluate the polynomials of degrees contained in ind at the points x.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The degrees of the polynomials to be evaluated.
        x : list or numpy.ndarray
            The points at which the polynomials are to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluation of the selected polynomials at the points x.

        '''
        x = np.ravel(x)
        coef = np.fliplr(self.poly_coeff(np.arange(np.max(ind)+1)))

        out = np.zeros((x.size, coef.shape[0]))
        for deg in range(coef.shape[0]):
            out[:, deg] = np.polyval(coef[deg, :], x)
        return out[:, ind]

    def d_poly_coef(self, ind):
        '''
        Return the coefficients of the first order derivative of the
        polynomials of degrees contained in ind.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The degrees of the polynomials to be evaluated.

        Returns
        -------
        numpy.ndarray
            The coefficients of the first order derivative of the polynomials
            of order contained in ind.

        '''
        return self.poly_coeff(ind)[:, 1:] * \
            np.tile(np.arange(1, np.max(ind)+1), (len(ind), 1))

    def dn_poly_coeff(self, n, ind):
        '''
        Return the coefficients of the n-th order derivative of the
        polynomials of degrees contained in ind.

        Parameters
        ----------
        n : int
            The degrees of derivation of the polynomials.
        ind : list or numpy.ndarray
            The orders of the polynomials to be evaluated.

        Returns
        -------
        numpy.ndarray
            The coefficients of the n-th order derivative of the polynomials
            of order contained in ind.

        '''
        coef = np.prod(np.tile(np.arange(np.max(ind)-n+1), (n, 1)) +
                       np.tile(np.expand_dims(np.arange(1, n+1), axis=1),
                               (1, np.max(ind)-n+1)), 0)
        coef = self.poly_coeff(ind)[:, n:] * np.tile(coef, (len(ind), 1))
        if coef.size == 0:
            coef = np.zeros(len(ind))
        return coef

    def d_polyval(self, ind, x):
        '''
        Evaluate the first order derivative of the polynomials of degrees
        contained in ind at the points x.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The degrees of the polynomials to be evaluated.
        x : list or numpy.ndarray
            The points at which the derivatives of the polynomials are to be
            evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluation of the first order derivatives of the selected
            polynomials at the points x.

        '''
        return self.dn_polyval(1, ind, x)

    def dn_polyval(self, n, ind, x):
        '''
        Evaluate the n-th order derivative of the polynomials of degrees
        contained in ind at the points x.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The degrees of the polynomials to be evaluated.
        x : list or numpy.ndarray
            The points at which the derivatives of the polynomials are to be
            evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluation of the n-th order derivatives of the selected
            polynomials at the points x.

        '''
        x = np.ravel(x)
        coef = np.fliplr(self.dn_poly_coeff(n, np.arange(np.max(ind)+1)))

        out = np.zeros((x.size, coef.shape[0]))
        for deg in range(coef.shape[0]):
            out[:, deg] = np.polyval(coef[deg, :], x)
        return out[:, ind]

    def mean(self, ind, measure=None):
        '''
        Return the mean of the polynomials of degrees contained in ind, with a
        Measure given by measure if provided, or to self.measure otherwise.

        Parameters
        ----------
        ind : list or numpy.ndarray
            The degrees of the polynomials for which the mean is to be
            computed.
        measure : tensap.Measure, optional
            The measure used for the computation of the mean. The default is
            None, indicating to use self.measure.

        Returns
        -------
        numpy.ndarray
            The mean of the selected polynomials.

        '''
        if measure is None:
            assert self.measure is not None, 'Must provide a Measure.'
            out = self.moment(np.reshape(ind, [-1, 1])) / self.measure.mass()
        else:
            out = self.moment(np.reshape(ind, [-1, 1]), measure) / \
                measure.mass()
        return out

    def moment(self, ind, measure=None):
        '''
        Return the integral of products of polynomials p_i, i in ind, using a gauss integration rule.
        
        The integral is with respect to a measure mu which is taken as the measure to the polynomials if not provided in input

        Assuming ind is a numpy.ndarray:
            - if ind.ndim == 1, and ind is of length N,  return the float
                m = int p_ind[0](x)...p_ind[N-1](x) dmu(x),
            - else if ind.ndim == 2, and ind is N-by-M,
                return the vector m of length N such that
                m[j] = int p_ind[j, 0](x)...p_ind[j, M-1](x) dmu(x)
                    

        Parameters
        ----------
        ind : list or numpy.ndarray
            Contains the degrees of the considered polynomials.
        measure : tensap.Measure, optional
            The measure used for the computation of the moments. The default is
            None, indicating to use self.measure.

        Returns
        -------
        numpy.ndarray
            Contains the integrals of products of polynomials.

        '''
        if measure is None:
            measure = self.measure

        ind = np.atleast_2d(ind)
        out = np.zeros(ind.shape[0])

        max_deg = np.max(np.sum(ind, 1))  # Maximum degree of the product
        nb_pts = int(np.ceil((max_deg+1)/2))  # Number of points

        G = measure.gauss_integration_rule(nb_pts)
        for i in range(ind.shape[0]):
            # Creation of the function to integrate
            def poly(x):
                out_loc = self.polyval(ind[i, 0], x)
                for j in np.arange(1, ind.shape[1]):
                    out_loc = out_loc * self.polyval(ind[i, j], x)
                return out_loc
            out[i] = G.integrate(poly)
        return out

    @staticmethod
    def ndim():
        '''
        Return the dimension of the output of the polynomials.

        Returns
        -------
        int
            The dimension of the output of the polynomials.

        '''
        return 1

    @abstractmethod
    def one(self):
        '''
        Coefficients and corresponding indices for the decomposition of the
        constant function 1.

        '''

    @abstractmethod
    def is_orthonormal(self):
        '''
        Check the orthonormality of the basis created by the functions of self.

        Returns
        -------
        bool
            Indicates if the polynomials are orthonormal.

        '''

    @abstractmethod
    def poly_coeff(self, ind):
        '''
        Compute the coefficients of the monomials used to create the
        polynomials of degree specified in ind.

        Parameters
        ----------
        ind : ind or numpy.ndarray
            The orders of the polynomials to be evaluated.

        '''


class CanonicalPolynomials(UnivariatePolynomials):
    '''
    Class CanonicalPolynomials.

    Attributes
    ----------
    measure : None or tensap.Measure
        The measure associated with the canonical polynomials.

    '''

    def __init__(self, measure=None):
        '''
        Constructor for the class CanonicalPolynomials.

        Parameters
        ----------
        measure : tensap.Measure, optional
            The measure associated with the canonical polynomials. The default
            is None.

        Raises
        ------
        ValueError
            If the provided measure is not a tensap.Measure object.

        Returns
        -------
        None.

        '''
        if measure is not None and not isinstance(measure, tensap.Measure):
            raise ValueError('Must provide a measure.')
        self.measure = measure

    def __eq__(self, poly_2):
        return isinstance(poly_2, CanonicalPolynomials)

    @staticmethod
    def is_orthonormal():
        return False

    def poly_coeff(self, ind):
        return np.eye(np.max(ind)+1)[ind, :]

    def d_polyval(self, ind, x):
        x = np.expand_dims(np.ravel(x), axis=1)
        ind = np.atleast_1d(ind)
        out = np.zeros((x.size, len(ind)))
        rep = ind != 0
        out[:, rep] = x**(ind[rep]-1) * np.tile(ind[rep], (x.size, 1))
        return out

    @staticmethod
    def domain():
        '''
        Return the domain of the canonical polynomials.

        Returns
        -------
        list
            The domain of the canonical polynomials.

        '''
        return [-np.inf, np.inf]

    @staticmethod
    def one():
        return 1, 0



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module normal_random_variable.

'''

from numpy import inf, array
from scipy.stats import norm
from numpy.random import randn
import tensap


class NormalRandomVariable(tensap.RandomVariable):
    '''
    Class NormalRandomVariable.

    Attributes
    ----------
    moments : numpy.array
        The moments of the normal random variable (if computed).
    mu : float, optional
        The mean of the normal random variable. The default is 0.
    sigma : float, optional
        The standard deviation of the normal random variable. The default is 1.

    '''

    def __init__(self, mean=0, standard_deviation=1):
        '''
        Normal random variable with provided mean and standard deviation.

        Parameters
        ----------
        mean : float, optional
            The mean of the normal random variable. The default is 0.
        standard_deviation : float, optional
            The standard deviation of the normal random variable. The default
            is 1.

        Returns
        -------
        None.

        '''
        tensap.RandomVariable.__init__(self)

        self.mu = mean
        self.sigma = standard_deviation

    def __repr__(self):
        return ('<{}: mu = {}, sigma = {}>').format(
            self.__class__.__name__,
            self.mu,
            self.sigma,
            t='\t', n='\n')

    def cdf(self, x):
        '''
        Evaluate the cumulative distribution function (cdf) of the normal
        random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the cdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the cdf.

        '''
        return norm.cdf(x, self.mu, self.sigma)

    def icdf(self, x):
        '''
        Evaluate the inverse cumulative distribution function (icdf) of the
        normal random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the icdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the icdf.

        '''
        return norm.ppf(x, self.mu, self.sigma)

    def pdf(self, x):
        '''
        Evaluate the probability density function (pdf) of the normal
        random variable at points x.

        Parameters
        ----------
        x : list or numpy.ndarray
            The points at which the pdf is to be evaluated.

        Returns
        -------
        numpy.ndarray
            The evaluations of the pdf.

        '''
        return norm.pdf(x, self.mu, self.sigma)

    def shift(self, bias, scaling):
        '''
        Shift the normal random variable using the provided bias and scaling
        factor.

        The mean mu of the random variable becomes mu + bias, and its standard
        deviation sigma becomes sigma * scaling.

        Parameters
        ----------
        bias : float
            The bias.
        scaling : float
            The scaling factor.

        Returns
        -------
        RV : tensap.NormalRandomVariable
            The shifted normal random variable.

        '''
        shifted_rv = tensap.NormalRandomVariable(self.mu, self.sigma)
        shifted_rv.mu += bias
        shifted_rv.sigma *= scaling
        return shifted_rv

    @staticmethod
    def get_standard_random_variable():
        '''
        Return the standard normal random variable with mean 0 and standard
        deviation 1.

        Returns
        -------
        tensap.NormalRandomVariable
            The standard normal random variable.

        '''
        return tensap.NormalRandomVariable()

    @staticmethod
    def support():
        '''
        Return the support of the normal random variable.

        Returns
        -------
        numpy.ndarray
            Support of the normal random variable.

        '''
        return array([-inf, inf])

    def orthonormal_polynomials(self, *max_degree):
        '''
        Return the max_degree-1 first orthonormal polynomials associated with
        the NormalRandomVariable.

        Parameters
        ----------
        max_degree : int, optional
            The maximum degree of the returned polynomials. The default is
            None, choosing the default maximum degree associated with the
            constructor of the polynomials.

        Returns
        -------
        poly : tensap.OrthonormalPolynomials
            The generated orthonormal polynomials.

        '''
        poly = tensap.HermitePolynomials(*max_degree)
        if self != NormalRandomVariable(0, 1):
            # print('ShiftedOrthonormalPolynomials are created.')
            poly = tensap.ShiftedOrthonormalPolynomials(poly, self.mu,
                                                        self.sigma)
        return poly

    def get_parameters(self):
        '''
        Return the parameters of the normal random variable.

        Returns
        -------
        float
            The mean of the random variable.
        float
            The standard deviation of the random variable.

        '''
        return self.mu, self.sigma

    def random_variable_statistics(self):
        return self.mu, self.sigma**2

    def random(self, n):
        return randn(int(n)) * self.sigma + self.mu



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module diagonal_tensor.

'''

from copy import deepcopy
import numpy as np
import tensap


class DiagonalTensor:
    '''
    Class DiagonalTensor.

    Attributes
    ----------
    data : numpy.ndarray
        The diagonal entries of the tensor.
    order : int
        The order of the tensor.
    shape : numpy.ndarray
        The shape of the tensor.
    is_orth : bool
        Boolean indicating if the representation of the tensor is orthogonal
        (i.e. one mu-matricization is orthogonal).

    '''

    def __init__(self, data, order=None):
        '''
        Constructor for the class DiagonalTensor.

        Parameters
        ----------
        data : numpy.ndarray or tensap.DiagonalTensor
            The diagonal entries of the tensor.
        order : int, optional
            The order of the tensor. The default is None.

        Returns
        -------
        None.

        '''
        self.is_orth = True

        if order is None:
            assert isinstance(data, DiagonalTensor), \
                'The input must be a DiagonalTensor.'
            self.data = np.array(data.data)
            self.order = data.order
            self.shape = data.shape
        else:
            assert isinstance(data, (list, np.ndarray)), \
                'The input must be a list or numpy.ndarray.'
            self.data = np.reshape(data, -1)
            self.order = order
            self.shape = np.full(order, self.data.size)

    def __repr__(self):
        return ('<{} DiagonalTensor:{n}' +
                '{t}order = {},{n}' +
                '{t}shape = {},{n}' +
                '{t}is_orth = {}>').format('x'.join(map(str, self.shape)),
                                           self.order,
                                           self.shape,
                                           self.is_orth,
                                           t='\t', n='\n')

    def tree_based_tensor(self, tree=None, is_active_node=None):
        '''
        Convert the tensap.DiagonalTensor into a tensap.TreeBasedTensor.

        Parameters
        ----------
        tree : tensap.DimensionTree, optional
            The tree associated with the tree-based tensor representation. The
            default is a linear tree.
        is_active_node : list or numpy.ndarray, optional
            List or array of booleans indicating if each node of the tree is
            active. The default is True for all nodes except the leaves.

        Raises
        ------
        ValueError
            If the internal nodes are not all active.

        Returns
        -------
        tensap.TreeBasedTensor
            A tree-based tensor representation of the diagonal tensor.

        '''
        if tree is None:
            tree = tensap.DimensionTree.linear(self.order)

        if is_active_node is None:
            is_active_node = np.full(tree.nb_nodes, True)
            is_active_node[tree.is_leaf] = False

        tensors = np.empty(tree.nb_nodes, dtype=object)
        tensors[np.logical_not(is_active_node)] = tensap.FullTensor([])
        r = self.shape[0]
        for nod in np.arange(1, tree.nb_nodes+1):
            ch = tree.children(nod)
            if tree.parent(nod) == 0:
                tensors[nod-1] = tensap.FullTensor.diag(self.data, ch.size)
            elif tree.is_leaf[nod-1] and is_active_node[nod-1]:
                tensors[nod-1] = tensap.FullTensor(np.eye(r), 2, [r, r])
            elif is_active_node[nod-1]:
                tensors[nod-1] = tensap.FullTensor.diag(np.ones(r), ch.size+1)
            elif not tree.is_leaf[nod-1] and not is_active_node[nod-1]:
                raise ValueError('The internal nodes should be active.')
        return tensap.TreeBasedTensor(tensors, tree)

    @property
    def ndim(self):
        '''
        Compute the order of the tensor. Equivalent to self.order.

        Returns
        -------
        int
            The order of the tensor.

        '''
        return self.order

    def storage(self):
        '''
        Return the storage complexity of the DiagonalTensor.

        Returns
        -------
        int
            The storage complexity of the DiagonalTensor.

        '''
        return self.data.size

    def sparse_storage(self):
        '''
        Return the sparse storage complexity of the DiagonalTensor.

        Returns
        -------
        int
            The sparse storage complexity of the DiagonalTensor.

        '''
        return np.count_nonzero(self.data)

    def __add__(self, y):
        return DiagonalTensor(self.data + y.data, self.order)

    def __neg__(self):
        return DiagonalTensor(-self.data, self.order)

    def __sub__(self, y):
        return DiagonalTensor(self.data - y.data, self.order)

    def __mul__(self, y):
        return DiagonalTensor(self.data * y.data, self.order)

    def reshape(self, shape):
        '''
        Reshape the tensor. The method has no effet.

        Parameters
        ----------
        shape : list or numpy.ndarray
            The new shape of the tensor.

        Returns
        -------
        tensor : DiagonalTensor
            The reshaped tensor.

        '''
        out = deepcopy(self)
        out.shape = np.ravel(shape)
        return out

    def sub_tensor(self, *args):
        '''
        Extract a subtensor of the tensor.

        See also tensap.FullTensor.sub_tensor.

        Parameters
        ----------
        *indices : list
            The indices to extract in each dimension. ':' indicates all the
            indices.

        Returns
        -------
        FullTensor
            The subtensor.

        '''

        out = self.full()
        return out.sub_tensor(*args)

    def update_attributes(self):
        '''
        Update the attribute shape of self if data or order have been modified.

        Returns
        -------
        None.

        '''
        self.shape = np.full(self.order, self.data.size)

    def tensor_vector_product(self, vectors):
        '''
        Compute the contraction of the tensor with vectors.

        Compute the contraction of self with each vector contained in the list
        vectors along all the dimensions. The operation is such that V[k] is
        contracted with the k-th dimension of self.

        Parameters
        ----------
        vectors : numpy.ndarray or list of numpy.ndarray
            The vectors to use in the product.

        Returns
        -------
        DiagonalTensor
            The tensor after the contractions with the vectors.

        '''
        if isinstance(vectors, list):
            vectors = np.hstack([np.reshape(x, [-1, 1]) for x in vectors])
        data = self.data * np.prod(vectors, 1)
        order = self.order - vectors.shape[1]
        if order == 0:
            return np.sum(data)
        return DiagonalTensor(data, order)

    def tensor_matrix_product(self, matrices, dims=None):
        '''
        Contract a tensor with matrices.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self, with the indices k given in dims (if provided).

        See also tensap.FullTensor.tensor_matrix_product.

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        FullTensor
            The tensor after the contractions with the matrices.

        '''
        out = self.full()
        return out.tensor_matrix_product(matrices, dims)

    def tensor_matrix_product_eval_diag(self, matrices, dims=None):
        '''
        Evaluate the diagonal of a tensor obtained by contraction with
        matrices.

        Provides the diagonal of the tensor obtained by contracting the tensor
        with matrices H[k] along dimensions dims(k)+1, for k = 0, ...,
        dims.size-1.

        Parameters
        ----------
        matrices : list
            The matrices to use in the product.
        dims : list or numpy.ndarray, optional
            Indices of the contractions. The default is None, indicating all
            the dimensions.

        Returns
        -------
        tensap.DiagonalTensor or numpy.ndarray
            The diagonal of the contractions of the tensor with the matrices.

        '''
        return self.tensor_matrix_product(matrices, dims).eval_diag(dims)

    def tensor_diagonal_matrix_product(self, matrices):
        '''
        Contract a FullTensor with matrices built from their diagonals.

        The second dimension of the matrix matrices[k] is contracted with the
        k-th dimension of self.

        Parameters
        ----------
        matrices : numpy.ndarray or list of numpy.ndarray
            The diagonals of the matrices to use in the product.

        Returns
        -------
        DiagonalTensor
            The tensor after the contractions with the matrices.

        '''
        if isinstance(matrices, list):
            matrices = np.hstack([np.reshape(x, [-1, 1]) for x in matrices])
        return DiagonalTensor(np.prod(matrices, 1) * self.data, self.order)

    def tensordot(self, y, dims1, dims2=None):
        '''
        Contract two tensors along specified dimensions.

        See also tensap.FullTensor.tensordot.

        Parameters
        ----------
        tensor2 : Tensor
            The second tensor of the contraction.
        dims1 : list or int
            The dimensions of contractions for the first tensor.
        dims2 : list or int, optional
            The dimensions of contractions for the second tensor. The default
            is None which indicates, if dims1 = 0, to perform the outer
            product of the two tensors, similarly to tensorflow.tensordot.

        Returns
        -------
        out : FullTensor
            The resulting tensor.

        '''
        x = self.full()
        y = y.full()
        return x.tensordot(y, dims1, dims2)

    def dot(self, y):
        '''
        Return the inner product of two tensors.

        Parameters
        ----------
        tensor2 : DiagonalTensor
            The second tensor of the inner products.

        Returns
        -------
        numpy.float
            The inner product of the two tensors.

        '''
        return np.sum(self.data*y.data)

    def norm(self):
        '''
        Compute the canonical norm of the DiagonalTensor.

        Returns
        -------
        numpy.float
            The norm of the tensor.

        '''
        return np.linalg.norm(self.data)

    def full(self):
        '''
        Convert the DiagonalTensor to a tensap.FullTensor.

        Returns
        -------
        tensap.FullTensor
            The DiagonalTensor as a tensap.FullTensor.

        '''
        return tensap.FullTensor.diag(self.data, self.order)

    def numpy(self):
        '''
        Convert the DiagonalTensor to a numpy.ndarray.

        Returns
        -------
        numpy.ndarray
            The DiagonalTensor as a numpy.ndarray.

        '''
        return self.full().data

    def sparse(self):
        '''
        Convert the DiagonalTensor to a tensap.SparseTensor.

        Returns
        -------
        tensap.SparseTensor
            The DiagonalTensor as a tensap.SparseTensor.

        '''
        ind = np.nonzero(self.data)[0]
        data = self.data[self.data != 0]
        indices = tensap.MultiIndices(np.tile(np.reshape(ind, [-1, 1]),
                                              (1, self.order)))
        return tensap.SparseTensor(data, indices, self.shape)

    def cat(self, y):
        '''
        Concatenate the tensors.

        Concatenates self and y in a tensor z such that:
        z(i_1 ,..., i_d) = x(i_1, ..., i_d) if i_k <= sz[k]-1 for k in dims,
        z(i_1, ..., i_d) = y(i_1-sz[0], ..., i_d-sz[d-1]) if i_k >= sz[k]
        for k in dims,
        z(i_1, ..., i_d) = 0 otherwise, with sz = self.shape and
        dims = range(self.order).

        Parameters
        ----------
        y : DiagonalTensor
            The second tensor to be concatenaed.
        Returns
        -------
        DiagonalTensor
            The concatenated tensors.

        '''
        data = np.concatenate((self.data, y.data))
        return DiagonalTensor(data, order=self.order)

    def kron(self, y):
        '''
        Kronecker product of tensors.

        Similar to numpy.kron but for arbitrary tensors.

        Parameters
        ----------
        tensor2 : DiagonalTensor
            The second tensor of the Kronecker product.

        Returns
        -------
        DiagonalTensor
            The tensor resulting from the Kronecker product.

        '''
        data = np.reshape(np.outer(y.data, self.data), -1, order='F')
        return DiagonalTensor(data, order=self.order)

    def dot_with_rank_one_metric(self, y, matrix):
        '''
        Compute the weighted inner product of two tensors.

        Compute the weighted canonical inner product of self and y,
        where the inner product related to dimension k is weighted by
        matrix[k]. It is equivalent to
        self.dot(y.tensor_matrix_product(matrix)), but can be much faster.

        Parameters
        ----------
        y : DiagonalTensor
            The second tensor of the inner product.
        matrix : list or numpy.ndarray or FullTensor
            The weight matrix.

        Returns
        -------
        numpy.float
            The weighted inner product.

        '''
        if isinstance(matrix, list):
            matrix = np.hstack([np.reshape(x, [-1, 1]) for x in matrix])
        matrix = np.reshape(np.expand_dims(matrix, 1),
                            (self.shape[0], y.shape[0], self.order), order='F')
        matrix = np.prod(matrix, 2)
        return np.matmul(np.transpose(self.data), np.matmul(matrix, y.data))

    def tensordot_matrix_product_except_dim(self, y, matrices, dim):
        '''
        Particular type of contraction.

        Compute a special contraction of two tensors self, y, a list of
        matrices matrices and a particular dimension dim. Note that dim must
        be a scalar, while matrices must be a list array with self.order
        elements.

        Parameters
        ----------
        y : DiagonalTensor
            The second tensor of the contraction.
        matrices : list
            The list of matrices of the contraction.
        dim : int
            The excluded dimension.

        Returns
        -------
        numpy.ndarray
            The result of the contraction.

        '''
        ind = np.setdiff1d(range(self.order), dim)
        matrices = np.hstack([matrices[x] for x in ind])
        matrices = np.reshape(np.expand_dims(matrices, 1),
                              (self.shape[0], y.shape[0], ind.size),
                              order='F')
        matrices = np.prod(matrices, 2)
        return matrices * np.outer(self.data, y.data)

    def orth(self):
        '''
        Placeholder method returning a copy of self.

        Returns
        -------
        DiagonalTensor
            A copy of self.

        '''
        return deepcopy(self)

    def eval_diag(self, dims=None):
        '''
        Extract the diagonal of the tensor.

        Parameters
        ----------
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices of the diagonal. The
            default is None,indicating that the indices refer to all the
            dimensions.

        Returns
        -------
        data : DiagonalTensor or numpy.ndarray
            The evaluations of the diagonal of the tensor.

        '''
        if dims is None or np.size(dims) == self.order:
            out = np.array(self.data)
        else:
            dims = np.sort(np.atleast_1d(dims))
            rep = np.concatenate((np.arange(dims[0]+1),
                                  np.arange(dims[-1]+1, self.order)))
            out = deepcopy(self)
            out.shape = out.shape[rep]
            out.order = np.size(rep)
        return out

    def eval_at_indices(self, ind, dims=None):
        '''
        Evaluate the tensor at indices.

        If dims is None, return
        s(k) = x(indices(k, 1), indices(k, 2), ..., indices(k, d)),
        1 <= k <= self.shape[0].

        If dims is not None, return a partial evaluation: up to a permutation
        (placing the dimensions dims on the left), return
        s(k, i_1, ..., i_d') = x(indices(k, 1), indices(k, 2), ...,
        indices(k, M), i_1, ..., i_d'),
        1 <= k <= self.shape[0], with M = dims.size and d' = self.order - M.

        Parameters
        ----------
        indices : list of numpy.ndarray
            The indices of the tensor.
        dims : list of numpy.ndarray, optional
            The dimensions associated with the indices. The default is None,
            indicating that indices refers to all the dimensions.

        Returns
        -------
        evaluations : numpy.ndarray or DiagonalTensor
            The evaluations of the tensor.

        '''

        ind = np.atleast_2d(ind)
        r = np.full(ind.shape[0], True)
        for k in range(ind.shape[1]):
            r = np.logical_and(r, ind[:, 0] == ind[:, k])

        if dims is None or np.size(dims) == self.order:
            out = np.zeros(ind.shape[0])
            out[r] = self.data[ind[r, 0]]
            return out
        raise NotImplementedError('Method not implemented.')

    def transpose(self, dims):
        '''
        Transpose (permute) the dimensions of the tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The new ordering of the dimensions.

        Returns
        -------
        tensor : DiagonalTensor
            The transposed (permuted) tensor.

        '''
        out = deepcopy(self)
        out.shape = out.shape[dims]
        return out

    def itranspose(self, dims):
        '''
        Return the inverse transpose (permutation) of the dimensions of the
        tensor.

        Parameters
        ----------
        dims : list or numpy.ndarray
            The original transpose (permutation) indices.

        Returns
        -------
        DiagonalTensor
            The transposed (permuted) tensor.

        '''
        return self.transpose(np.argsort(dims))

    @staticmethod
    def create(generator, rank, order):
        '''
        Create a DiagonalTensor of rank rank and order order using a given
        generator.

        Parameters
        ----------
        generator : function
            Function generating an array, given a rank.
        rank : int
            The rank of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        DiagonalTensor
            The created tensor.

        '''
        return DiagonalTensor(generator(rank), order)

    @staticmethod
    def rand(rank, order):
        '''
        Create a DiagonalTensor of rank rank and order order with i.i.d.
        entries drawn according to the uniform distribution on [0, 1].

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        DiagonalTensor
            The created tensor.

        '''
        return DiagonalTensor.create(np.random.rand, rank, order)

    @staticmethod
    def randn(rank, order):
        '''
        Create a DiagonalTensor of rank rank and order order with i.i.d.
        entries drawn according to the standard gaussian distribution.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        DiagonalTensor
            The created tensor.

        '''
        return DiagonalTensor.create(np.random.randn, rank, order)

    @staticmethod
    def zeros(rank, order):
        '''
        Create a DiagonalTensor of rank rank and order order with with entries
        equal to 0.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        DiagonalTensor
            The created tensor.

        '''
        return DiagonalTensor.create(np.zeros, rank, order)

    @staticmethod
    def ones(rank, order):
        '''
        Create a DiagonalTensor of rank rank and order order with with
        entries equal to 1.

        Parameters
        ----------
        rank : int
            The rank of the tensor.
        order : int
            The order of the tensor.

        Returns
        -------
        DiagonalTensor
            The created tensor.

        '''
        return DiagonalTensor.create(np.ones, rank, order)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module user_defined_function.

'''

import numpy as np
import tensap


class UserDefinedFunction(tensap.Function):
    '''
    Class UserDefinedFunction.

    Attributes
    ----------
    fun : function or tensap.Function
        The function defining the UserDefinedFunction object.

    '''

    def __init__(self, fun, dim, shape=None):
        '''
        Constructor for the class UserDefinedFunction.

        Parameters
        ----------
        fun : function or tensap.Function
            The function defining the UserDefinedFunction object.
        dim : int
            The number of input variables.
        shape : list or numpy.ndarray, optional
            The shape of the output of the function. The default is 1.

        Returns
        -------
        None.

        '''
        tensap.Function.__init__(self)

        self.evaluation_at_multiple_points = False
        self.dim = dim
        if shape is not None:
            self.output_shape = shape

        if isinstance(fun, str):
            for i in np.arange(dim-1, -1, -1):
                s = f'x{i}'
                s_new = f'x[:, {i}]'
                fun = fun.replace(s, s_new)
            self.fun = lambda x: eval(fun)
        else:
            self.fun = fun

    def eval(self, x):
        if np.ndim(x) == 1:
            x = np.expand_dims(x, 1)

        n = x.shape[0]

        if hasattr(self.fun, 'eval'):
            def fun(x):
                return self.fun.eval(x)
        else:
            fun = self.fun

        if self.evaluation_at_multiple_points:
            y = fun(x)
            if isinstance(y, list):
                y = np.hstack([np.reshape(y_loc, [-1, 1]) for y_loc in y])
        else:
            y = np.zeros((n, np.prod(self.output_shape)))
            for i in range(n):
                y[i, :] = np.ravel(fun(np.atleast_2d(x[i, :])))

        return np.reshape(y, np.concatenate(
            ([n], np.atleast_1d(self.output_shape))))



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module linear_model_learning_square_loss.

'''

import numpy as np
from scipy.sparse import diags
import tensap


class LinearModelLearningSquareLoss(tensap.LinearModelLearning):
    '''
    Class LinearModelLearningSquareLoss.

    Attributes
    ----------
    weights : list or numpy.ndarray, optional
        The arrays for the weighted least-squares minimization. The default is
        None.
    linear_solver : str
        The selected solver: 'solve' to directly solve the normal equation
        using numpy.solve, 'qr' to perform a QR decomposition, and solve
        the resulting system using numpy.solve. The default is 'qr'.
    shared_coefficients : bool
        When approximating a vector-valued function, indicates if the
        coefficients of the approximation are common to all the outputs. The
        default is True.

    '''

    def __init__(self, weights=None):
        '''
        Constructor for the class LinearModelLearningSquareLoss.

        When approximating a vector-valued function, setting sharedCoefficients
        to false will independently compute y.shape[1] sets of coefficients,
        whereas setting it to true will compute one set of coefficients shared
        across all the outputs. In that case, basisEval should be an array of
        shape (n-by-N-by-D), with n the size of the dataset, N the number of
        basis functions and D the number of outputs.

        Parameters
        ----------
        weights : list or numpy.ndarray, optional
            The arrays for the weighted least-squares minimization. The default
            is None.

        Returns
        -------
        None.

        '''
        super().__init__(tensap.SquareLossFunction())
        self.weights = weights
        self.linear_solver = 'qr'
        self.shared_coefficients = True

    def solve(self):
        '''
        Solution (Ordinary or Regularized) of the Least-Squares problem and
        cross-validation procedure.

        Returns
        -------
        sol : numpy.ndarray or tensap.FunctionalBasisArray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        self.initialize()

        A = self.basis_eval
        y = self.training_data[1]

        if self.shared_coefficients:
            if np.ndim(A) == 3:
                assert (np.ndim(y) == 1 and A.shape[2] == 1) or \
                    (np.ndim(y) == 2 and A.shape[2] == y.shape[1]), \
                    'A.shape[2] should be equal to y.shape[1].'

                A = np.transpose(A, [0, 2, 1])
                A = np.reshape(A, [-1, A.shape[2]], order='F')
                y = np.reshape(y, [-1, 1], order='F')
                self.basis_eval = A
                self.training_data[1] = y
        if np.ndim(y) == 2:
            n = y.shape[1]
        else:
            n = 1

        output = {}

        if self.weights is not None:
            weights = diags(np.sqrt(self.weights))
            A = np.matmul(weights, A)
            y = np.matmul(weights, y)

        if not self.basis_adaptation:
            if not self.regularization:
                sol, output = self._solve_ols()
            else:
                if n == 1:
                    sol, output = self._solve_regularized_ls()
                else:
                    sol = np.zeros([A.shape[1], n])
                    output['error'] = np.zeros(n)
                    output['outputs'] = np.empty(n, dtype=object)
                    for ind in range(n):
                        self.training_data[1] = y[:, ind]
                        sol[:, ind], output_tmp = self._solve_regularized_ls()
                        output['error'][ind] = output_tmp['error']
                        output['outputs'][ind] = output_tmp
        else:
            if n == 1:
                sol, output = self._solve_basis_adaptation()
            else:
                sol = np.zeros([A.shape[1], n])
                output['error'] = np.zeros(n)
                output['outputs'] = np.empty(n, dtype=object)
                for ind in range(n):
                    self.training_data[1] = y[:, ind]
                    sol[:, ind], output_tmp = self._solve_basis_adaptation()
                    output['error'][ind] = output_tmp['error']
                    output['outputs'][ind] = output_tmp

        if self.test_error:
            if n == 1:
                f_eval = np.matmul(self.basis_eval_test, sol)
                output['test_error'] = self.loss_function.test_error(
                    f_eval, self.test_data)
            else:
                output['test_error'] = np.zeros([1, n])
                f_eval = np.matmul(self.basis_eval_test, sol)
                for ind in range(n):
                    test_data = [self.test_data[0], self.test_data[1][:, ind]]
                    output['test_error'][ind] = self.loss_function.test_error(
                        f_eval, test_data)

        if self.basis is not None:
            sol = tensap.FunctionalBasisArray(sol, self.basis, n)

        return sol, output

    def _solve_ols(self):
        '''
        Solution of the Ordinary Least-Squares (OLS) problem and
        cross-validation procedure (if requested).

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        A = self.basis_eval
        y = self.training_data[1]
        output = {}

        if self.linear_solver == 'solve':
            B = np.matmul(np.transpose(A), A)
            try:
                sol = np.linalg.solve(B, np.matmul(np.transpose(A), y))
            except Exception:
                sol = np.linalg.lstsq(B, np.matmul(np.transpose(A), y),
                                      None)[0]
        elif self.linear_solver == 'qr':
            q_A, r_A = np.linalg.qr(A)
            try:
                sol = np.linalg.solve(r_A, np.matmul(np.transpose(q_A), y))
            except Exception:
                sol = np.linalg.lstsq(r_A, np.matmul(np.transpose(q_A), y),
                                      None)[0]

        if self.error_estimation:
            second_moment = np.var(y, 0) + np.mean(y, 0)**2
            if self.error_estimation_type == 'residuals':
                delta = y - np.matmul(A, sol)
                error = np.mean(delta**2, 0) / second_moment
            else:
                if self.linear_solver == 'solve':
                    try:
                        B_inv = np.linalg.inv(B)
                    except Exception:
                        B_inv = np.linalg.pinv(B)
                    error, delta = self._compute_cv_error(y, second_moment,
                                                          A, sol,
                                                          self.linear_solver,
                                                          B_inv)
                elif self.linear_solver == 'qr':
                    error, delta = self._compute_cv_error(y, second_moment,
                                                          A, sol,
                                                          self.linear_solver,
                                                          q_A, r_A)
            output['error'] = error
            output['delta'] = delta

        output['flag'] = 2
        return sol, output

    def _solve_regularized_ls(self):
        '''
        Solution of the Regularized Least-Squares problem and cross-validation
        procedure.

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        from sklearn import linear_model

        A = self.basis_eval
        y = np.squeeze(self.training_data[1])
        output = {}

        N, P = A.shape
        solpath = []

        options = self.regularization_options
        if self.regularization_type == 'l0':
            options_loc = {i: output[i] for i in output if i != 'alpha'}
            D = np.linalg.norm(A, axis=0)
            A_omp = np.matmul(A, diags(1/D))
            solpath = linear_model.orthogonal_mp(A_omp, y,
                                                 copy_X=True,
                                                 n_nonzero_coefs=np.min((N,
                                                                         P)),
                                                 return_path=True,
                                                 **options_loc)
            solpath = np.atleast_2d(solpath)
            sol = np.matmul(diags(1/D), solpath[:, -1])
        elif self.regularization_type == 'l1':
            reg = linear_model.LassoLars(copy_X=True, fit_intercept=False,
                                         **options)
            reg.fit(A, y)
            sol = reg.coef_
            solpath = reg.coef_path_
        elif self.regularization_type == 'l2':
            reg = linear_model.Ridge(copy_X=True, fit_intercept=False,
                                     **options)
            reg.fit(A, y)
            sol = reg.coef_
        else:
            raise ValueError('Regularization technique not implemented.')

        if self.model_selection and np.linalg.norm(solpath) != 0:
            if 'non_zero_blocks' in self.options:
                rep = np.true(solpath.shape[1])
                rep = np.logical_and(rep, np.any(solpath, 0))
                for block in self.options['non_zero_blocks']:
                    rep = np.logical_and(rep, np.any(solpath[block, :], 0))
                solpath = solpath[:, rep]

            sol, output = self._select_optimal_path(solpath)
        elif self.model_selection:
            # print('solpath does not exist or is empty.')
            pass

        if self.error_estimation and 'error' not in output:
            second_moment = np.var(y, 0) + np.mean(y, 0)**2
            if self.error_estimation_type == 'residuals':
                delta = y - np.matmul(A, sol)
                error = np.mean(delta**2, 0) / second_moment
            else:
                ind = np.nonzero(sol)[0]
                A_red = A[:, ind]
                if self.linear_solver == 'solve':
                    B = np.matmul(np.transpose(A_red), A_red)
                    try:
                        B_inv = np.linalg.inv(B)
                    except Exception:
                        B_inv = np.linalg.pinv(B)
                    error, delta = self._compute_cv_error(y, second_moment,
                                                          A_red, sol[ind],
                                                          self.linear_solver,
                                                          B_inv)
                elif self.linear_solver == 'qr':
                    q_A, r_A = np.linalg.qr(A_red)
                    error, delta = self._compute_cv_error(y, second_moment,
                                                          A_red, sol[ind],
                                                          self.linear_solver,
                                                          q_A, r_A)
            output['error'] = error
            output['delta'] = delta

        output['flag'] = 2
        return sol, output

    def _solve_basis_adaptation(self):
        '''
        Solution of the Least-Squares problem with working-set and
        cross-validation procedure.

        Returns
        -------
        sol : numpy.ndarray
            The solution of the minimization problem.
        output : dict
            Outputs of the algorithm.

        '''
        if self.basis_adaptation_path is None:
            solpath = np.triu(np.full([self.basis_eval.shape[1]]*2, True))
        else:
            solpath = self.basis_adaptation_path

        sol, output = self._select_optimal_path(solpath)
        output['flag'] = 2
        return sol, output

    def _select_optimal_path(self, solpath=None):
        '''
        Selection of a solution using (corrected) relative (leave-one-out or
        k-fold) cross-validation error.

        The dictionnary output contains:
            - error: the (corrected) relative (leave-one-out or k-fold)
              cross-validation error estimate,
            - error_path: the regularization path of (corrected) relative
              (leave-one-out or k-fold) cross-validation error estimates,
            - ind: the index of the optimal pattern,
            - pattern: the optimal sparsity pattern,
            - pattern_path: the regularization path of sparsity patterns,
            - solution_path: the regularization path of potential non-zero
              solutions with different sparsity patterns,
            - solution_path_OLS: the regularization path of the OLS solutions,
            - optimal_solution: the optimal solution,
            - delta: the optimal residual,
            - delta_path: the regularization paths of the residuals.

        Parameters
        ----------
        solpath : numpy.ndarray, optional
            Array of shape (P, m) whose columns give m potential solutions with
            different sparsity patterns. The default is None, returning the
            solution of the OLS problem without basis adaptation.

        Returns
        -------
        numpy.ndarray
            The optimal solution of the minimization problem.
        dict
            Outputs of the algorithm.

        '''
        if solpath is None:
            return self._solve_ols()

        A = self.basis_eval
        y = self.training_data[1]
        output = {}
        linear_solver = self.linear_solver

        pattern = solpath != 0
        rep = np.any(pattern, 0)
        pattern = pattern[:, rep]
        solpath = solpath[:, rep]

        _, ind = np.unique(pattern, return_index=True, axis=1)
        ind = np.sort(ind)
        solpath = solpath[:, ind]
        pattern = pattern[:, ind]

        sol = np.zeros(pattern.shape)

        if 'gram_matrix' in self.error_estimation_options:
            gram_matrix = self.error_estimation_options['gram_matrix']

        second_moment = np.var(y, 0) + np.mean(y, 0)**2

        ind = np.sum(pattern != 0, 0) <= A.shape[0]
        err = np.zeros(np.count_nonzero(ind))
        delta = np.zeros([y.shape[0], np.count_nonzero(ind)])
        pattern = pattern[:, ind]
        solpath = solpath[:, ind]

        if np.linalg.norm(solpath) == 0:
            return self._solve_ols()

        for i in range(pattern.shape[1]):
            ind = pattern[:, i]

            A_red = A[:, ind]
            if 'gram_matrix' in self.error_estimation_options:
                self.error_estimation_options['gram_matrix'] =  \
                    gram_matrix[ind, ind]

            if A_red.shape[1] > A_red.shape[0]:
                self.linear_solver = 'solve'
            else:
                self.linear_solver = linear_solver

            if self.linear_solver == 'qr':
                q_A, r_A = np.linalg.qr(A_red)
                try:
                    sol_red = np.linalg.solve(r_A,
                                              np.matmul(np.transpose(q_A), y))
                except Exception:
                    sol_red = np.linalg.lstsq(r_A,
                                              np.matmul(np.transpose(q_A), y),
                                              None)[0]

                err[i], delta[:, i] = self._compute_cv_error(y, second_moment,
                                                             A_red, sol_red,
                                                             'qr', q_A, r_A)
            elif self.linear_solver == 'solve':
                try:
                    C_red = np.linalg.inv(np.matmul(np.transpose(A_red),
                                                    A_red))
                except Exception:
                    C_red = np.linalg.pinv(np.matmul(np.transpose(A_red),
                                                     A_red))

                sol_red = np.matmul(C_red, np.matmul(np.transpose(A_red), y))
                err[i], delta[:, i] = self._compute_cv_error(y, second_moment,
                                                             A_red, sol_red,
                                                             'solve', C_red)

            sol[ind, i] = np.squeeze(sol_red)

            if i > 1 and \
                self.model_selection_options['stop_if_error_increase'] and \
                    err[i] > 2*err[i-1]:
                print('stop_if_error_increase')
                err[i+1:] = np.inf
                delta[:, i+1:] = np.inf

        ind = np.argmin(err)

        output['error'] = err[ind]
        output['error_path'] = err
        output['ind'] = ind
        output['pattern'] = pattern[:, ind]
        output['pattern_path'] = pattern
        output['solution_path'] = solpath
        output['optimal_solution'] = sol[:, ind]
        output['solution_path_OLS'] = sol
        output['delta'] = delta[:, ind]
        output['delta_path'] = delta

        return sol[:, ind], output

    def _compute_cv_error(self, y, second_moment, A, sol, linear_solver=None,
                          *args):
        '''
        Relative cross-validation error estimates and residuals.

        For self.error_estimation_type = 'leave_out': compute the relative
        leave-one-out cross-validation error for the coefficients matrix sol
        using the fast leave-one-out cross-validation procedure
        [Cawlet & Talbot 2004] based on the Bartlett matrix inversion formula
        (special case of the Sherman-Morrison-Woodbury formula).
        If self.error_estimation_type.correction == True, compute the corrected
        relative leave-one-out cross-validation error for the coefficients
        matrix sol.

        For self.error_estimation_type = 'k_fold': compute the  k-fold
        cross-validation error for the coefficients matrix sol using the fast
        k-fold cross-validation procedure based on the
        Sherman-Morrison-Woodbury formula.
        self.error_estimation_options['number_of_folds']: number of folds
        (only for the k-fold cross-validation procedure), min(10, N) by default
        where N is the number of samples.
        If self.error_estimation_type.correction == True, compute the corrected
        relative k-fold cross-validation error for the coefficients matrix sol.


        Parameters
        ----------
        y : numpy.ndarray
            The evaluations of response vector.
        second_moment : float or numpy.ndarray
            The empirical second moment of y.
        A : numpy.ndarray
            The evaluations of the basis functions.
        sol : numpy.ndarray
            The coefficients of the approximation.
        linear_solver : str, optional
            The selected linear solver. The default is None, selecting 'solve'.
        *args : tuple
            Additional arguments (if required).

        Raises
        ------
        NotImplementedError
            If the requested cross-validation estimator is not implemented.

        Returns
        -------
        numpy.ndarray
            The (corrected) relative (leave-one-out or k-fold) cross-validation
            error estimate.
        delta : numpy.ndarray
            The residuals.

        '''
        if np.ndim(y) == 2:
            N, n = y.shape
        else:
            N = y.shape[0]
            n = 1

        P = A.shape[1]
        err = np.zeros(n)
        delta = np.zeros(y.shape)

        if self.error_estimation_type == 'k_fold':
            if 'number_of_folds' not in self.error_estimation_options:
                self.error_estimation_options['number_of_folds'] = np.min([10,
                                                                           N])
            if self.error_estimation_options['number_of_folds'] == N:
                self.error_estimation_type = 'leave_out'

        if not args and linear_solver is None:
            if P > N:
                try:
                    C = np.linalg.inv(np.matmul(np.transpose(A), A))
                except Exception:
                    C = np.linalg.pinv(np.matmul(np.transpose(A), A))
                linear_solver = 'solve'
            else:
                q_A, r_A = np.linalg.qr(A)
                linear_solver = 'qr'
        else:
            if linear_solver == 'qr':
                q_A, r_A = args
            elif linear_solver == 'solve':
                C = args[0]

        # Compute the absolute cross-validation error (cross-validation error
        # estimate of the mean-squared error), also called mean predicted
        # residual sum of squares (PRESS) or empirical mean-squared predicted
        # residual
        if self.error_estimation_type == 'leave_out':
            if N-1 < P:
                err = np.full(err.shape, np.inf)
                delta = np.squeeze(np.full(delta.shape, np.inf))
                if isinstance(err, np.ndarray) and err.size == 1:
                    err = err[0]
                return err, delta

            # Compute the predicted residuals using the Bartlett matrix
            # inversion formula (special case of the Sherman-Morrison-Woodbury
            # formula)
            if linear_solver == 'solve':
                T = np.sum(np.transpose(A) * np.matmul(C, np.transpose(A)), 0)
            elif linear_solver == 'qr':
                T = np.sum(q_A**2, 1)

            with np.errstate(divide='ignore', invalid='ignore'):
                delta = (y - np.matmul(A, sol)) / (1-np.reshape(T, y.shape))
            delta = np.squeeze(delta)
            # Compute the absolute cross-validation error
            err = np.mean(delta**2, 0)
        elif self.error_estimation_type == 'k_fold':
            from sklearn.model_selection import KFold

            n_folds = np.min(
                (N, self.error_estimation_options['number_of_folds']))

            if N * (1 - 1 / n_folds) < P:
                # print('Not enough samples for performing OLS on the ' +
                #       'training set.')
                err = np.full(err.shape, np.inf)
                delta = np.squeeze(np.full(delta.shape, np.inf))
                if isinstance(err, np.ndarray) and err.size == 1:
                    err = err[0]
                return err, delta

            cvp = KFold(n_splits=n_folds, shuffle=True).split(A)

            errors = []
            delta = np.zeros(y.shape)

            if linear_solver == 'solve':
                H = np.matmul(A, np.matmul(C, np.transpose(A)))
            elif linear_solver == 'qr':
                H = np.matmul(q_A, np.transpose(q_A))

            for _, test in cvp:
                # Compute the predicted residual for the current fold using the
                # Sherman-Morrison-Woodbury formula
                if n == 1:
                    y_loc = y[test]
                else:
                    y_loc = y[test, :]
                delta_loc = np.linalg.solve(np.eye(len(test)) -
                                            H[np.ix_(test, test)],
                                            y_loc-np.matmul(A[test, :], sol))
                # Compute the absolute cross-validation error for the current
                # fold
                errors.append(np.mean(delta_loc**2))
                if n == 1:
                    delta[test] = delta_loc
                else:
                    delta[test, :] = delta_loc

            # Average over the folds
            err = np.mean(errors)
            delta = np.squeeze(delta)
        else:
            raise NotImplementedError('Cross-validation {} not implemented'.
                                      format(self.error_estimation_type))

        # Compute the relative cross-validation error
        err = err / second_moment

        # Compute the corrected relative cross-validation error to reduce the
        # sensitivity of the error estimate to overfitting (the non-corrected
        # cross-validation error estimate underpredicts the error in L^2-norm
        # (generalization error))
        if linear_solver == 'qr' and \
                1 / np.linalg.cond(r_A) < np.finfo(float).eps:
            pass
        elif self.error_estimation_options['correction']:
            if N != P:
                if linear_solver == 'qr':
                    try:
                        inv_rA = np.linalg.inv(r_A)
                    except Exception:
                        inv_rA = np.linalg.pinv(r_A)
                    C = np.matmul(inv_rA, np.transpose(inv_rA))

                if 'gram_matrix' in self.error_estimation_options:
                    # Direct Eigenvalue Estimator (DEE) in the case of non
                    # orthonormal bases [Chapelle, Vapnik & Bengio, 2002],
                    # [Blatman & Sudret, 2011].
                    # The Gram matrix must be provided. Independence between
                    # variables is assumed
                    corr = (N / (N-P)) * (1 + np.trace(np.matmul(
                        C, self.error_estimation_options['gram_matrix'])))
                else:
                    # Direct Eigenvalue Estimator
                    # [Chapelle, Vapnik & Bengio, 2002],
                    # [Blatman & Sudret, 2011] -> accurate even when N is
                    # not >> P (corr ~ 1+2*P/N when N->Inf, as trace(C) ~ P/N
                    # when N->Inf)
                    corr = (N / (N-P)) * (1 + np.trace(C))

                if corr > 0:
                    err *= corr

        if isinstance(err, np.ndarray) and err.size == 1:
            err = err[0]
        return np.sqrt(err), delta



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Tutorial on digits recognition using tree-based tensor formats.

'''

from sklearn import datasets, metrics
import random
import numpy as np
import tensorflow as tf
import time
import matplotlib.pyplot as plt
import tensap

# %% Data import and preparation
DIGITS = datasets.load_digits()
DATA = DIGITS.images.reshape((len(DIGITS.images), -1))
DATA = DATA / np.max(DATA)  # Scaling of the data

# %% Patch reshape of the data: the patches are consecutive entries of the data
PS = [4, 4]  # Patch size
DATA = np.array([np.concatenate(
    [np.ravel(np.reshape(DATA[k, :], [8]*2)[PS[0]*i:PS[0]*i+PS[0],
                                            PS[1]*j:PS[1]*j+PS[1]]) for
     i in range(int(8/PS[0])) for j in range(int(8/PS[1]))]) for
    k in range(DATA.shape[0])])
DIM = int(int(DATA.shape[1]/np.prod(PS)))

# %% Probability measure
print('Dimension %i' % DIM)
X = tensap.RandomVector(tensap.DiscreteRandomVariable(np.unique(DATA)), DIM)

# %% Training and test samples
P_TRAIN = 0.9  # Proportion of the sample used for the training

N = DATA.shape[0]
TRAIN = random.sample(range(N), int(np.round(P_TRAIN*N)))
TEST = np.setdiff1d(range(N), TRAIN)
X_TRAIN = DATA[TRAIN, :]
X_TEST = DATA[TEST, :]
Y_TRAIN = DIGITS.target[TRAIN]
Y_TEST = DIGITS.target[TEST]

# One hot encoding (vector-valued function)
Y_TRAIN = tf.one_hot(Y_TRAIN.astype(int), 10, dtype=tf.float64)
Y_TEST = tf.one_hot(Y_TEST.astype(int), 10, dtype=tf.float64)

# %% Approximation bases: 1, cos and sin for each pixel of the patch
FUN = [lambda x: np.ones((np.shape(x)[0], 1))]
for i in range(np.prod(PS)):
    FUN.append(lambda x, j=i: np.cos(np.pi / 2*x[:, j]))
    FUN.append(lambda x, j=i: np.sin(np.pi / 2*x[:, j]))

BASES = [tensap.UserDefinedFunctionalBasis(FUN, X.random_variables[0],
                                           np.prod(PS)) for _ in range(DIM)]
BASES = tensap.FunctionalBases(BASES)

# %% Loss function: cross-entropy custom loss function
LOSS = tensap.CustomLossFunction(
        lambda y_true, y_pred: tf.nn.sigmoid_cross_entropy_with_logits(
            logits=y_pred, labels=y_true))


def error_function(y_pred, sample):
    '''
    Return the error associated with a set of predictions using a sample, equal
    to the number of misclassifications divided by the number of samples.

    Parameters
    ----------
    y_pred : numpy.ndarray
        The predictions.
    sample : list
        The sample used to compute the error. sample[0] contains the inputs,
        and sample[1] the corresponding outputs.

    Returns
    -------
    int
        The error.

    '''
    try:
        y_pred = y_pred(sample[0])
    except Exception:
        pass
    return np.count_nonzero(np.argmax(y_pred, 1) - np.argmax(sample[1], 1)) / \
        sample[1].numpy().shape[0]


LOSS.error_function = error_function

# %% Learning in tree-based tensor format
TREE = tensap.DimensionTree.balanced(DIM)
IS_ACTIVE_NODE = np.full(TREE.nb_nodes, True)
SOLVER = tensap.TreeBasedTensorLearning(TREE, IS_ACTIVE_NODE, LOSS)

SOLVER.tolerance['on_stagnation'] = 1e-10
SOLVER.initialization_type = 'random'
SOLVER.bases = BASES
SOLVER.training_data = [X_TRAIN, Y_TRAIN]
SOLVER.test_error = True
SOLVER.test_data = [X_TEST, Y_TEST]

SOLVER.rank_adaptation = True
SOLVER.rank_adaptation_options['max_iterations'] = 15
SOLVER.model_selection = True
SOLVER.display = True

SOLVER.alternating_minimization_parameters['display'] = False
SOLVER.alternating_minimization_parameters['max_iterations'] = 10
SOLVER.alternating_minimization_parameters['stagnation'] = 1e-10

# Options dedicated to the LinearModelCustomLoss object
SOLVER.linear_model_learning.options['max_iterations'] = 10
SOLVER.linear_model_learning.options['stagnation'] = 1e-10
SOLVER.linear_model_learning.optimizer.learning_rate = 1e3

SOLVER.rank_adaptation_options['early_stopping'] = True
SOLVER.rank_adaptation_options['early_stopping_factor'] = 10

T0 = time.time()
F, OUTPUT = SOLVER.solve()
T1 = time.time()
print(T1-T0)

# %% Display of the results
F_X_TEST = np.argmax(F(X_TEST), 1)
Y_TEST_NP = np.argmax(Y_TEST.numpy(), 1)

print('\nAccuracy = %2.5e\n' % (1 - np.count_nonzero(F_X_TEST - Y_TEST_NP) /
                                Y_TEST_NP.shape[0]))

IMAGES_AND_PREDICTIONS = list(zip(DIGITS.images[TEST], F_X_TEST))
for i in np.arange(1, 19):
    plt.subplot(3, 6, i)
    plt.imshow(IMAGES_AND_PREDICTIONS[i][0],
               cmap=plt.cm.gray_r, interpolation='nearest')
    plt.axis('off')
    plt.title('Pred.: %i' % IMAGES_AND_PREDICTIONS[i][1])

print('Classification report:\n%s\n'
      % (metrics.classification_report(Y_TEST_NP, F_X_TEST)))
MATRIX = metrics.confusion_matrix(Y_TEST_NP, F_X_TEST)
plt.matshow(MATRIX)
plt.title('Confusion Matrix')
plt.show()
print('Confusion matrix:\n%s' % MATRIX)



# Copyright (c) 2020, Anthony Nouy, Erwan Grelier
# This file is part of tensap (tensor approximation package).

# tensap is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tensap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with tensap.  If not, see <https://www.gnu.org/licenses/>.

'''
Module multi_indices.

'''

import numpy as np
# import tensap


class MultiIndices:
    '''
    Class MultiIndices.

    Attributes
    ----------
    array : numpy.ndarray
        An array containing n multi-indices in N^d.

    '''

    def __init__(self, array=None):
        '''
        Constructor for the class MultiIndices.

        Parameters
        ----------
        array : list or numpy.ndarray or tensap.MultiIndices, optional
            The array characterizing the set of multi-indices. The default is
            None.

        Returns
        -------
        None.

        '''
        if hasattr(array, 'array'):
            array = array.array
        self.array = np.atleast_2d(array)

    def sub2ind(self, shape):
        '''
        Convert the indices of the MultiIndices object into flat indices.

        Parameters
        ----------
        shape : list or numpy.ndarray
            The shape of the array.

        Returns
        -------
        numpy.ndarray
            The flat indices associated with the MultiIndices object and shape.

        '''
        ind = self.to_list()
        return np.ravel_multi_index(ind, shape, order='F')

    def __eq__(self, J):
        if not isinstance(J, MultiIndices):
            ok = False
        else:
            if J.cardinal() == 1:
                J.array = np.tile(J.array, (self.cardinal(), 1))
            ok = np.all(self.array == J.array, axis=1)
        return ok

    def __le__(self, J):
        assert isinstance(J, MultiIndices), 'Must provide a MultiIndices.'
        if J.cardinal() == 1:
            J.array = np.tile(J.array, (self.cardinal(), 1))
        return np.all(self.array <= J.array, axis=1)

    def __add__(self, m):
        return MultiIndices(self.array+m)

    def __radd__(self, m):
        self.array +=m

    def __sub__(self, m):
        return MultiIndices(self.array-m)

    def __rsub__(self, m):
        self.array -=m

    def cardinal(self):
        '''
        Return the cardinal of the MultiIndices object.

        Returns
        -------
        int
            The cardinal of the MultiIndices object.

        '''
        return self.array.shape[0]

    def to_list(self):
        '''
        Convert the MultiIndices' array into a list of arrays.

        Returns
        -------
        list
            The MultiIndices' array as a list of arrays.

        '''
        return [self.array[:, i] for i in range(self.ndim())]

    def ndim(self):
        '''
        Return the dimension of the multi-indices.

        Returns
        -------
        int
            The dimension of the multi-indices.

        '''
        return self.array.shape[1]

    def norm(self, p=2, k=None):
        '''
        Compute the p-norm of multi-indices k in the object.

        Parameters
        ----------
        p : int or numpy.inf, optional
            The positive real scalar p of the p-norm, or numpy.inf. The default
            is 2.
        k : list or numpy.ndarray, optional
            The multi-indices of which the norm is to be computed. The default
            is all the multi-indices of the object.

        Returns
        -------
        norm : numpy.ndarray
            The p-norm of the selected multi-indices.

        '''
        if k is None:
            k = np.arange(self.cardinal())
        if p == np.inf:
            norm = np.max(self.array[k, :], axis=1)
        else:
            norm = np.power(np.sum(self.array[k, :]**p, axis=1), 1/p)
        return norm

    def weighted_norm(self, p, w, k=None):
        '''
        Compute the weighted p-norm of multi-indices k in the object.

        Parameters
        ----------
        p : int or numpy.inf
            The positive real scalar p of the p-norm, or numpy.inf.
        w : list or numpy.ndarray
            The self.cardinal() weights used in the computation of the norm.
        k : list or numpy.ndarray, optional
            The multi-indices of which the norm is to be computed. The default
            is all the multi-indices of the object.

        Returns
        -------
        norm : numpy.ndarray
            The p-norm of the selected multi-indices.

        '''
        if k is None:
            k = np.arange(self.cardinal())
        if p == np.inf:
            norm = np.max(self.array[k, :]*np.tile(np.ravel(w),
                                                   (np.size(k), 1)), axis=1)
        else:
            norm = np.power(np.sum((self.array[k, :] *
                                    np.tile(np.ravel(w), (np.size(k), 1)))**p,
                                   axis=1), 1/p)
        return norm

    def sort_by_norm(self, p, mode='ascend'):
        '''
        Sort the multi-indices by increasing or decreasing p-norm.

        Parameters
        ----------
        p : int
            The positive real scalar p of the p-norm, or numpy.inf.
        mode : string, optional
            The sorting mode, equal to 'ascend' or 'descend'. The default is
            'ascend'.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices object with multi-indices sorted by increasing or
            decreasing p-norm.

        '''
        norm = self.norm(p)
        ind = np.argsort(norm)
        if mode == 'descend':
            ind = np.flip(ind)
        return MultiIndices(self.array[ind, :])

    def sort_by_weighted_norm(self, p, w, mode='ascend'):
        '''
        Sort the multi-indices by increasing or decreasing weighted p-norm.

        Parameters
        ----------
        p : int
            The positive real scalar p of the p-norm, or numpy.inf.
        w : list or numpy.ndarray
            The self.cardinal() weights used in the computation of the norm.
        mode : string, optional
            The sorting mode, equal to 'ascend' or 'descend'. The default is
            'ascend'.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices object with multi-indices sorted by increasing or
            decreasing weighted p-norm.

        '''
        norm = self.weighted_norm(p, w)
        ind = np.argsort(norm)
        if mode == 'descend':
            ind = np.flip(ind)
        return MultiIndices(self.array[ind, :])

    def sort(self, columns=None, mode='ascend'):
        '''
        Sort multi-indices column-wise using the column order provided in
        columns.

        The method first sorts according to the column columns[0], then sorts
        the equal coefficients of column columns[1] and so on.

        Parameters
        ----------
        columns : list or numpy.ndarray, optional
            The column order used to sort the multi-indices. The default is
            last to first column.
        mode : string, optional
            The sorting mode, equal to 'ascend' or 'descend'. The default is
            'ascend'.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices object with sorted multi-indices.

        '''
        if columns is None:
            columns = np.arange(self.array.shape[1]-1, -1, -1)

        array = np.array(self.array)
        for k in np.flip(columns):
            if k == columns[-1]:
                ind = np.argsort(array[:, k])
            else:
                ind = np.argsort(array[:, k], kind='mergesort')
            if mode == 'descend':
                ind = np.flip(ind)
            array = array[ind, :]
        return MultiIndices(array.astype(int))

    def add_indices(self, J):
        '''
        Return the union of multi-indices of self and J.

        Parameters
        ----------
        J : tensap.MultiIndices
            The second MultiIndices object.

        Returns
        -------
        tensap.MultiIndices
            The union of multi-indices of self and J.

        '''
        array = np.vstack((self.array, J.array))
        ind = np.unique(array, axis=0, return_index=True)[1]
        array = np.array([array[index, :] for index in sorted(ind)])
        return MultiIndices(array.astype(int))

    def remove_indices(self, J):
        '''
        Remove multi-indices J from self.

        Parameters
        ----------
        J : tensap.MultiIndices or int or list or numpy.ndarray
            The multi-indices to remove in a tensap.MultiIndices object,
            or their numbers as an int or in a list or numpy.ndarray.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices object with removed multi-indices in J.

        '''
        if isinstance(J, MultiIndices):
            ind = np.nonzero(np.all(self.array == J.array[:, np.newaxis],
                                    axis=2))[1]
        else:
            ind = J
        array = self.array[np.setdiff1d(range(self.cardinal()), ind), :]
        return MultiIndices(array.astype(int))

    def remove_dims(self, dims):
        '''
        Remove the dimensions in ind in the MultiIndices.

        Parameters
        ----------
        dims : int or list or numpy.ndarray
            The dimension(s) to be removed.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices object with removed dimenions in dims.

        '''
        array = self.array[:, np.setdiff1d(range(self.ndim()), dims)]
        ind = np.unique(array, axis=0, return_index=True)[1]
        array = np.array([array[index, :] for index in sorted(ind)])
        return MultiIndices(array.astype(int))

    def intersect_indices(self, J):
        '''
        Return the intersection of the multi-indices of self and J.

        Parameters
        ----------
        J : tensap.MultiIndices
            The second multi-indices.

        Returns
        -------
        tensap.MultiIndices
            The intersection of the multi-indices of self and J.
        ind_I : numpy.array
            The indices of the multi-indices in self common to self and J.
        ind_J : numpy.array
            The indices of the multi-indices in J common to self and J.

        '''
        ind_J, ind_I = np.nonzero(np.all(self.array == J.array[:, np.newaxis],
                                         axis=2))

        return MultiIndices(self.array[ind_I, :]), ind_I, ind_J

    def keep_indices(self, k):
        '''
        Keep the multi-indices k in self.

        Parameters
        ----------
        k : int or list or numpy.ndarray
            The number(s) of the multi-indices to keep.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices with retained indices.

        '''
        return MultiIndices(self.array[k, :])

    def keep_dims(self, dims):
        '''
        Keep the dimensions dims in self.

        Parameters
        ----------
        dims : int or list or numpy.ndarray
            The dimension(s) to keep.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices with retained dimensions.

        '''
        array = self.array[:, dims]
        ind = np.unique(array, axis=0, return_index=True)[1]
        array = np.array([array[index, :] for index in sorted(ind)])
        return MultiIndices(array.astype(int))

    def get_indices(self, k):
        '''
        Return the multi-indices k in self.

        Parameters
        ----------
        k : int or list or numpy.ndarray
            The numbers of the multi-indices to select.

        Returns
        -------
        numpy.ndarray
            The multi-indices k in self.

        '''
        return self.array[k, :]

    def is_downward_closed(self, m=0):
        '''
        Check whether or not the multi-index set is downward closed (or lower
        or monotone).

        Parameters
        ----------
        m : int, optional
            The lowest index for all dimensions. The default is 0.

        Returns
        -------
        cond : boolean
            Boolean indicating whether or not the multi-index set is downward
            closed.

        '''
        cond = True
        ind_test = np.arange(self.cardinal())
        while ind_test.size:
            p = self.array[ind_test[-1], :]
            Ip = MultiIndices.bounded_by(p, m)
            ind = np.all(Ip.array == self.array[:, np.newaxis], axis=2)
            ok = np.any(ind, axis=0)
            rep = np.nonzero(ind)[0]
            if not np.all(ok):
                cond = False
                return cond
            ind_test = np.setdiff1d(ind_test, rep)
        return cond

    def envelope(self, u):
        '''
        Compute the monotone envelope (or monotone majorant) of a bounded
        sequence u.

        Parameters
        ----------
        u : list or numpy.ndarray
            The bounded sequence.

        Returns
        -------
        env : numpy.ndarray
            The monotone envelope corresponding to the sequence defined by
            :math:`env_i = max_{j >= i} |u_j|`

        '''
        array = self.array
        n = self.cardinal()
        assert np.size(u) == n, \
            ('The length of the sequence does not coincide with the number ' +
             'of multi-indices.')

        env = np.array(u)
        for i in range(n):
            ind_sup = np.all(array >= np.tile(array[i, :], (n, 1)), axis=1)
            env[i] = np.max(np.abs(u[ind_sup]))
        return env

    def get_maximal_indices(self):
        '''
        Return the set of maximal multi-indices contained in the downward
        closed multi-index set self.

        Returns
        -------
        tensap.MultiIndices
            The set of maximal multi-indices contained in the downward closed
            multi-index set self.

        '''
        dim = self.ndim()
        n = self.cardinal()
        neighbours = np.tile(np.transpose(np.expand_dims(self.array, 2),
                                          [0, 2, 1]), [1, dim, 1]) + \
            np.tile(np.transpose(np.expand_dims(np.eye(dim), 2),
                                 [2, 0, 1]), [n, 1, 1])
        neighbours = np.reshape(neighbours, [n*dim, dim],
                                order='F').astype(int)
        ok = np.any(np.all(neighbours == self.array[:, np.newaxis], axis=2),
                    axis=0)
        ok = np.reshape(ok, [n, dim], order='F')
        ind_max = self.array[np.logical_not(np.any(ok, axis=1)), :]
        return MultiIndices(ind_max.astype(int))

    def get_margin(self):
        '''
        Return the margin of the multi-index set self defined by the set of
        multi-indices i not in self such that it exists k in N^* s.t. i_k != 0
        implies i - e_k in self where e_k is the k-th Kronecker sequence.

        Returns
        -------
        tensap.MultiIndices
            The margin of self.

        '''
        dim = self.ndim()
        n = self.cardinal()
        neighbours = np.tile(np.transpose(np.expand_dims(self.array, 2),
                                          [0, 2, 1]), [1, dim, 1]) + \
            np.tile(np.transpose(np.expand_dims(np.eye(dim), 2),
                                 [2, 0, 1]), [n, 1, 1])
        neighbours = np.reshape(neighbours, [n*dim, dim],
                                order='F').astype(int)

        ind_marg = np.nonzero(np.all(neighbours == self.array[:, np.newaxis],
                                     axis=2))[1]
        ind_marg = neighbours[np.setdiff1d(range(neighbours.shape[0]),
                                           ind_marg), :]

        ind = np.unique(ind_marg, axis=0, return_index=True)[1]
        ind_marg = np.array([ind_marg[index, :] for index in sorted(ind)])

        return MultiIndices(ind_marg.astype(int))

    def get_reduced_margin(self):
        '''
        Return the reduced margin of the multi-index set self defined by the
        set of multi-indices i not in self such that for all k in N^* s.t.
        i_k != 0 implies i - e_k in self where e_k is the k-th Kronecker
        sequence.

        Returns
        -------
        tensap.MultiIndices
            The reduced margin of self.

        '''
        I_marg = self.get_margin()
        dim = self.ndim()
        neighbours = np.tile(np.transpose(np.expand_dims(I_marg.array, 2),
                                          [0, 2, 1]), [1, dim, 1]) - \
            np.tile(np.transpose(np.expand_dims(np.eye(dim), 2), [2, 0, 1]),
                    [I_marg.cardinal(), 1, 1])

        n = neighbours.shape[0]
        neighbours = np.reshape(neighbours, [n*dim, dim],
                                order='F').astype(int)

        ok = np.any(np.all(neighbours == self.array[:, np.newaxis], axis=2),
                    axis=0)
        is_out = np.any(neighbours < 0, axis=1)
        ok = np.logical_or(ok, is_out)
        ok = np.reshape(ok, [n, dim], order='F')
        keep = np.all(ok, axis=1)
        ind_marg_red = I_marg.array[keep, :]

        return MultiIndices(ind_marg_red.astype(int))

    def plot(self, *args):
        '''
        Plot the multi-index set self.

        See also the function plot_multi_indices.

        Parameters
        ----------
        *args : tuple
            Parameters used in tensap's function plot_multi_indices.

        Returns
        -------
        None.

        '''
        # TODO plot
        # tensap.plot_multi_indices(self, *args)

    @staticmethod
    def with_bounded_norm(d, p, m):
        '''
        Create the set of multi-indices in N^d with p-norm bounded by m, p>0.

        Parameters
        ----------
        d : int
            The dimension, a positive integer.
        p : int or numpy.inf
            The p of the p-norm, either a positive real scalar or numpy.inf.
        m : int
            The bound of the norm, a positive real scalar.

        Returns
        -------
        ind : tensap.MultiIndices
            The set of multi-indices in N^d with p-norm bounded by m.

        '''
        if p == np.inf:
            ind = MultiIndices.product_set(np.arange(m+1), d)
        elif p == 1:
            ind = MultiIndices(np.zeros(d, dtype=int))
            for _ in range(m):
                ind = ind.add_indices(ind.get_margin())
        else:
            ind = MultiIndices(np.zeros(d, dtype=int))
            add = True
            while add:
                M = ind.get_margin()
                n = M.norm(p)
                k = np.nonzero(n <= m)[0]
                if np.all(np.logical_not(k)):
                    add = False
                else:
                    j = np.argsort(n[k])
                    M = M.keep_indices(k[j])
                    ind = ind.add_indices(M)
        return ind

    @staticmethod
    def with_bounded_weighted_norm(d, p, m, w):
        '''
        Create the set of multi-indices in N^d with weighted p-norm bounded by
        m, p>0.

        Parameters
        ----------
        d : int
            The dimension, a positive integer.
        p : int or numpy.inf
            The p of the p-norm, either a positive real scalar or numpy.inf.
        m : int
            The bound of the norm, a positive real scalar.
        w : list or numpy.ndarray
            The d weights defining the weighted p-norm of a multi-index i.
            :math:`|i|_{p, w} = (sum_{k=1}^d w_k^p i_k^p)^(1/p) for 0 < p < inf`
            :math:`|i|_{Inf, w} = max_k w_k i_k`

        Returns
        -------
        ind : tensap.MultiIndices
            The set of multi-indices in N^d with weighted p-norm bounded by m.

        '''
        ind = [np.arange(int(np.floor(m/x))+1) for x in w]
        ind = MultiIndices.product_set(ind)
        n = ind.weighted_norm(p, w)
        k = np.nonzero(n <= m)[0]
        j = np.argsort(n[k])
        return ind.keep_indices(k[j])

    @staticmethod
    def bounded_by(m, m0=0):
        '''
        Create the set of multi-indices bounded by m.

        Parameters
        ----------
        m : list or numpy.ndarray
            List or array of length d containing the highest indices in each
            dimension.
        m0 : int, optional
            The lowest index for all dimensions. The default is 0.

        Returns
        -------
        tensap.MultiIndices
            MultiIndices object containing the  product set
            (m0:m[0]) x ... x (m0:m[d-1]).

        '''
        return MultiIndices.product_set([np.arange(m0, x+1) for x in m])

    @staticmethod
    def product_set(L, d=None):
        '''
        Create the set of multi-indices obtained by a product of sets of
        indices.

        If ndim(L) is 1 and L contains a set of indices, the method returns
        a MultiIndices containing the product set L x ... x L (d times).

        If ndim(L) is 2 and L contains arrays of integers of size m_k,
        0 <= k <= d-1, the method returns a MultiIndices containing the product
        set L[0] x ... x L[d-1].

        Parameters
        ----------
        L : list or numpy.ndarray
            The grid or grids used to create the product set.
        d : int, optional
            The dimension of the set of multi-indices. The default is None,
            indicating to infer it from L.

        Raises
        ------
        ValueError
            If the provided arguments are wrong.

        Returns
        -------
        tensap.MultiIndices
            The set of multi-indices obtained by a product of sets of indices.

        '''
        if d is None:
            d = len(L)
        elif np.ndim(L) == 1:
            L = [L]*d
        else:
            raise ValueError('Wrong arguments.')

        L = [np.array(x) for x in L]
        N = [x.size for x in L]

        ind = list(np.unravel_index(range(np.prod(N)), N, order='F'))

        for i in range(d):
            ind[i] = L[i][ind[i]]

        return MultiIndices(np.transpose(np.array(ind)))

    @staticmethod
    def ind2sub(shape, ind):
        '''
        Create the set of multi-indices with array [I1, ..., Id] such that
        (I1, ..., Id) = np.unravel_index(np.ravel(ind), shape).

        Parameters
        ----------
        shape : list or numpy.ndarray
            The shape of the array.
        ind : list or numpy.ndarray
            The indices into the flattened version of an array of dimensions
            shape.

        Returns
        -------
        tensap.MultiIndices
            The MultiIndices created using the flat indices ind and shape.

        '''
        ind = np.unravel_index(np.ravel(ind), shape, order='F')
        return MultiIndices(np.transpose(ind))
